import argparse
import collections
import json
import os


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("card_db_src_dir")
    parser.add_argument("--write", action="store_true")
    args = parser.parse_args()

    with open(os.path.join(args.card_db_src_dir, "en_us/cards_en_us.json")) as f:
        en_contents = json.load(f)

    for fname in os.listdir(args.card_db_src_dir):
        if fname not in ["en_us", "xx"] and os.path.isdir(
            os.path.join(args.card_db_src_dir, fname)
        ):
            print(f"processing {fname}")
            full_fname = os.path.join(
                args.card_db_src_dir, fname, f"cards_{fname}.json"
            )
            with open(full_fname) as f:
                other_contents = json.load(f)
            trimmed = {}
            trimmed_counts = collections.Counter()

            for card_name, card_spec in other_contents.items():
                matched = en_contents.get(card_name)
                if matched is None:
                    print(f"warning - leaving {card_name} as it doesn't exist in en_us")
                    trimmed[card_name] = card_spec
                    continue
                new_card_spec = {}
                for key in ["description", "extra", "name"]:
                    entry = card_spec.get(key)
                    if entry is not None and entry != matched[key]:
                        new_card_spec[key] = entry
                    else:
                        trimmed_counts[key] += 1
                # do not remove name if something else is translated
                if (
                    "name" in card_spec
                    and "name" not in new_card_spec
                    and len(set(new_card_spec) - {"name"}) > 0
                ):
                    new_card_spec["name"] = card_spec["name"]
                    trimmed_counts["name"] -= 1
                if new_card_spec:
                    # preserve other keys
                    for other_key in set(card_spec) - {"name", "description", "extra"}:
                        new_card_spec[other_key] = card_spec[other_key]
                    trimmed[card_name] = new_card_spec

            print(f"trimmed: {dict(trimmed_counts)}")
            print(f"{len(trimmed)} cards left of {len(other_contents)}")
            print(f"{len(trimmed)} cards (partially) translated of {len(en_contents)}")

            if args.write:
                with open(full_fname, "wt", encoding="utf-8") as outf:
                    json.dump(trimmed, outf, indent=4, ensure_ascii=False)
                    outf.write("\n")
                print("rewrote contents")

            print()


if __name__ == "__main__":
    main()



import fnmatch
import json
import os
import sys
import unicodedata
from collections import Counter, defaultdict

from loguru import logger
from reportlab.lib.units import cm

from . import config_options, db
from .cards import Card
from .draw import DividerDrawer

try:
    from icu import Collator, Locale

    have_icu = True
except ImportError:
    have_icu = False


def generate_sample(options):
    from io import BytesIO

    from wand.image import Image

    buf = BytesIO()
    options.num_pages = 1
    options.outfile = buf
    generate(options)
    sample_out = BytesIO()
    with Image(blob=buf.getvalue(), resolution=options.preview_resolution) as sample:
        sample.format = "png"
        sample.save(sample_out)
        return sample_out.getvalue()


class CardSorter(object):
    def __init__(self, order, lang, baseCards):
        self.order = order

        # If PyICU has been successfully imported
        if have_icu:
            # Create a sort collator based on the selected language. Will be used the generate the sort keys.
            self.collator = Collator.createInstance(Locale(lang))
        else:
            logger.warning(
                "PyICU library not found. The dividers will be ordered by default sort key (might not be the "
                "correct alphabetical order for the selected language)."
            )

            self.collator = None

        if order == "global":
            self.sort_key = self.by_global_sort_key
        elif order == "colour":
            self.sort_key = self.by_colour_sort_key
        elif order == "cost":
            self.sort_key = self.by_cost_sort_key
        else:
            self.sort_key = self.by_expansion_sort_key

        self.baseOrder = [
            "Copper",
            "Silver",
            "Gold",
            "Platinum",
            "Potion",
            "Curse",
            "Estate",
            "Duchy",
            "Province",
            "Colony",
            "Trash",
            "Start Deck",
        ]
        self.baseCards = []
        for tag in self.baseOrder:
            if tag in baseCards:
                self.baseCards.append(baseCards[tag])
                del baseCards[tag]
        # now pick up those that have not been specified
        for tag in baseCards:
            self.baseCards.append(baseCards[tag])

    # When sorting cards, want to always put "base" cards after all
    # kingdom cards, and order the base cards in a particular order
    # (ie, all normal treasures by worth, then potion, then all
    # normal VP cards by worth, then Trash)
    def baseIndex(self, name):
        try:
            return self.baseCards.index(name)
        except Exception:
            return -1

    def isBaseExpansionCard(self, card):
        return card.cardset_tag.lower() != "base" and card.name in self.baseCards

    def by_global_sort_key(self, card):
        return (
            int(card.isExpansion()),
            self.baseIndex(card.name),
            self.get_card_name_sort_key(card.name),
        )

    def by_expansion_sort_key(self, card):
        return (
            card.cardset,
            int(card.isExpansion()),
            self.baseIndex(card.name),
            self.get_card_name_sort_key(card.name),
        )

    def by_colour_sort_key(self, card):
        return card.getType().getTypeNames(), self.get_card_name_sort_key(card.name)

    def by_cost_sort_key(self, card):
        return (
            card.cardset,
            int(card.isExpansion()),
            str(card.get_total_cost(card)),
            self.get_card_name_sort_key(card.name),
        )

    def get_card_name_sort_key(self, c):
        if self.collator:  # If the PyICU collator attribute has been set up, get the collator based sort key
            return self.collator.getSortKey(c)
        else:  # Default method: strip the card name character accents
            return self.strip_accents(c)

    @staticmethod
    def strip_accents(s):
        return "".join(
            c
            for c in unicodedata.normalize("NFD", s)
            if unicodedata.category(c) != "Mn"
        )

    def __call__(self, card):
        return self.sort_key(card)


def add_card_text(cards, language="en_us"):
    language = language.lower()
    # Read in the card text file
    card_text_filepath = os.path.join(
        "card_db", language, "cards_" + language.lower() + ".json.gz"
    )
    with db.get_resource_stream(card_text_filepath) as card_text_file:
        card_text = json.loads(card_text_file.read().decode("utf-8"))
    assert language, "Could not load card text for %r" % language

    # Now apply to all the cards
    for card in cards:
        if card.card_tag in card_text:
            if "name" in card_text[card.card_tag].keys():
                card.name = card_text[card.card_tag]["name"]
            if "description" in card_text[card.card_tag].keys():
                card.description = card_text[card.card_tag]["description"]
            if "extra" in card_text[card.card_tag].keys():
                card.extra = card_text[card.card_tag]["extra"]
    return cards


def add_set_text(options, sets, language="en_us"):
    language = language.lower()
    # Read in the set text and store for later
    set_text_filepath = os.path.join("card_db", language, f"sets_{language}.json.gz")
    with db.get_resource_stream(set_text_filepath) as set_text_file:
        set_text = json.loads(set_text_file.read().decode("utf-8"))
    assert set_text, "Could not load set text for %r" % language

    # Now apply to all the sets
    for s in sets:
        if s in set_text:
            for key in set_text[s]:
                sets[s][key] = set_text[s][key]
    return sets


def add_type_text(types=None, language="en_us"):
    if types is None:
        types = {}
    language = language.lower()
    # Read in the type text and store for later
    type_text_filepath = os.path.join("card_db", language, f"types_{language}.json.gz")
    with db.get_resource_stream(type_text_filepath) as type_text_file:
        type_text = json.loads(type_text_file.read().decode("utf-8"))
    assert type_text, "Could not load type text for %r" % language

    # Now apply to all the types
    used = {}
    for type in types:
        if type in type_text:
            types[type] = type_text[type]
            used[type] = True

    for type in type_text:
        if type not in used:
            types[type] = type_text[type]

    return types


def add_bonus_regex(options, language="en_us"):
    language = language.lower()
    # Read in the bonus regex terms
    bonus_regex_filepath = os.path.join(
        "card_db", language, f"bonuses_{language}.json.gz"
    )
    with db.get_resource_stream(bonus_regex_filepath) as bonus_regex_file:
        bonus_regex = json.loads(bonus_regex_file.read().decode("utf-8"))
    assert bonus_regex, "Could not load bonus keywords for %r" % language

    if not bonus_regex:
        bonus_regex = {}

    return bonus_regex


def combine_cards(cards, old_card_type, new_card_tag, new_cardset_tag, new_type):
    holder = Card(
        name="*Replace Later*",
        card_tag=new_card_tag,
        group_tag=new_card_tag,
        cardset_tag=new_cardset_tag,
        types=(new_type,),
        count=0,
    )
    holder.image = holder.setImage()

    filteredCards = []
    for c in cards:
        if c.isType(old_card_type):
            holder.addCardCount(c.count)  # keep track of count and skip card
        else:
            filteredCards.append(c)  # Not the right type, keep card

    if holder.getCardCount() > 0:
        filteredCards.append(holder)

    return filteredCards


def filter_sort_cards(cards, options):
    # Filter out cards by edition
    if options.edition and options.edition != "all":
        keep_sets = []
        for set_tag in Card.sets:
            for edition in Card.sets[set_tag]["edition"]:
                if options.edition == edition:
                    keep_sets.append(set_tag)

        keep_cards = []  # holds the cards that are to be kept
        for card in cards:
            if card.cardset_tag in keep_sets:
                keep_cards.append(card)

        cards = keep_cards

    # Combine upgrade cards with their expansion
    if options.upgrade_with_expansion:
        if options.exclude_expansions is None:
            options.exclude_expansions = []
        for card in cards:
            if Card.sets[card.cardset_tag]["upgrades"]:
                options.exclude_expansions.append(card.cardset_tag.lower())
                card.cardset_tag = Card.sets[card.cardset_tag]["upgrades"]

    # Combine globally all cards of the given types
    # For example, Events, Landmarks, Projects, Ways, Traits
    if options.group_global:
        # First find all possible types to group that match options.group_global
        types_to_group = {}
        for t in Card.types:
            group_global_type = Card.types[t].getGroupGlobalType()
            if group_global_type:
                theType = "-".join(t)
                # Save if either the old or the new type matches the option
                # Remember options.global_group is already lowercase
                if theType.lower() in options.group_global:
                    types_to_group[theType] = group_global_type
                elif group_global_type.lower() in options.group_global:
                    types_to_group[theType] = group_global_type

        # Now work through the matching types to group
        for t in types_to_group:
            cards = combine_cards(
                cards,
                old_card_type=t,
                new_type=types_to_group[t],
                new_card_tag=types_to_group[t].lower(),
                new_cardset_tag=config_options.EXPANSION_GLOBAL_GROUP,
            )
        if options.expansions:
            options.expansions.append(config_options.EXPANSION_GLOBAL_GROUP)

    # Take care of any blank cards
    if options.include_blanks > 0:
        if options.expansions:
            options.expansions.append(config_options.EXPANSION_GLOBAL_GROUP)

    # Group all the special cards together
    if options.group_special:
        keep_cards = []  # holds the cards that are to be kept
        group_cards = {}  # holds the cards for each group
        for card in cards:
            if not card.group_tag:
                keep_cards.append(card)  # not part of a group, so just keep the card
            else:
                # have a card in a group
                if (card.group_tag, card.cardset_tag) not in group_cards:
                    # First card of a group
                    group_cards[(card.group_tag, card.cardset_tag)] = (
                        card  # save to update cost later
                    )
                    # this card becomes the card holder for the whole group.
                    card.card_tag = card.group_tag
                    # These text fields should be updated later if there is a translation for this group_tag.
                    error_msg = (
                        "ERROR: Missing language entry for group_tab '%s'."
                        % card.group_tag
                    )
                    card.name = (
                        card.group_tag
                    )  # For now, change the name to the group_tab
                    card.description = error_msg
                    card.extra = error_msg
                    if card.get_GroupCost():
                        card.cost = card.get_GroupCost()
                    # now save the card
                    keep_cards.append(card)
                else:
                    # subsequent cards in the group. Update group info, but don't keep the card.
                    if card.group_top:
                        # this is a designated card to represent the group, so update important data
                        group_cards[(card.group_tag, card.cardset_tag)].cost = card.cost
                        group_cards[
                            (card.group_tag, card.cardset_tag)
                        ].potcost = card.potcost
                        group_cards[
                            (card.group_tag, card.cardset_tag)
                        ].debtcost = card.debtcost
                        group_cards[
                            (card.group_tag, card.cardset_tag)
                        ].types = card.types
                        group_cards[
                            (card.group_tag, card.cardset_tag)
                        ].randomizer = card.randomizer
                        group_cards[
                            (card.group_tag, card.cardset_tag)
                        ].image = card.image

                    group_cards[(card.group_tag, card.cardset_tag)].addCardCount(
                        card.count
                    )  # increase the count
                    # set holder to lowest cost of the two cards
                    # group_cards[(card.group_tag, card.cardset_tag)].set_lowest_cost(card)

        cards = keep_cards

        # Now fix up card costs for groups by Type (Events, Landmarks, etc.)
        for card in cards:
            if (card.card_tag, card.cardset_tag) in group_cards and group_cards[
                (card.group_tag, card.cardset_tag)
            ].get_GroupCost():
                group_cards[(card.group_tag, card.cardset_tag)].cost = group_cards[
                    (card.group_tag, card.cardset_tag)
                ].get_GroupCost()
                group_cards[(card.group_tag, card.cardset_tag)].debtcost = 0
                group_cards[(card.group_tag, card.cardset_tag)].potcost = 0

    # Get the final type names in the requested language
    Card.type_names = add_type_text(Card.type_names, db.LANGUAGE_DEFAULT)
    if options.language != db.LANGUAGE_DEFAULT:
        Card.type_names = add_type_text(Card.type_names, options.language)
    for card in cards:
        card.types_name = " - ".join([Card.type_names[t] for t in card.types])

    # Get the card bonus keywords in the requested language
    bonus = add_bonus_regex(options, db.LANGUAGE_DEFAULT)
    Card.addBonusRegex(bonus)
    if options.language != db.LANGUAGE_DEFAULT:
        bonus = add_bonus_regex(options, options.language)
        Card.addBonusRegex(bonus)

    # Fix up cardset text.  Waited as long as possible.
    Card.sets = add_set_text(options, Card.sets, db.LANGUAGE_DEFAULT)
    if options.language != db.LANGUAGE_DEFAULT:
        Card.sets = add_set_text(options, Card.sets, options.language)

    # Split out Official and Fan set information
    Official_sets = set()  # Will hold official sets
    Official_search = []  # Will hold official sets for searching, both set key and set_name
    Fan_sets = set()  # Will hold fan sets
    Fan_search = []  # Will hold fan sets for searching, both set key and set_name
    wantedSets = set()  # Will hold all the sets requested for printing

    All_search = []  # Will hold all sets for searching, both set key and set_name
    for s in Card.sets:
        search_items = [s.lower(), Card.sets[s].get("set_name", None).lower()]
        All_search.extend(search_items)
        if Card.sets[s].get("fan", False):
            # Fan Expansion
            Fan_sets.add(s)
            Fan_search.extend(search_items)
        else:
            # Official Expansion
            Official_sets.add(s)
            Official_search.extend(search_items)

    # If expansion names given, then find out which expansions are requested
    # Expansion names can be the names from the language or the cardset_tag
    if options.expansions:
        # Expand out any wildcards, matching set key or set name in the given language
        expanded_expansions = []
        for e in options.expansions:
            matches = fnmatch.filter(Official_search, e)
            if matches:
                expanded_expansions.extend(matches)
            else:
                expanded_expansions.append(e)

        # Now get the actual sets that are matched above
        options.expansions = set(
            [e.lower() for e in expanded_expansions]
        )  # Remove duplicates
        knownExpansions = set()
        for e in options.expansions:
            for s in Official_sets:
                if s.lower() == e or Card.sets[s].get("set_name", "").lower() == e:
                    wantedSets.add(s)
                    knownExpansions.add(e)
        # Give indication if an imput did not match anything
        unknownExpansions = options.expansions - knownExpansions
        if unknownExpansions:
            logger.warning((f"Unknown expansion(s): {', '.join(unknownExpansions)}"))

    # Take care of fan expansions.  Fan expansions must be explicitly named to be added.
    # If no --fan is given, then no fan cards are added.
    # Fan expansion names can be the names from the language or the cardset_tag
    if options.fan:
        # Expand out any wildcards, matching set key or set name in the given language
        expanded_expansions = []
        for e in options.fan:
            matches = fnmatch.filter(Fan_search, e)
            if matches:
                expanded_expansions.extend(matches)
            else:
                expanded_expansions.append(e)

        # Now get the actual sets that are matched above
        options.fan = set([e.lower() for e in expanded_expansions])  # Remove duplicates
        knownExpansions = set()
        for e in options.fan:
            for s in Fan_sets:
                if s.lower() == e or Card.sets[s].get("set_name", "").lower() == e:
                    wantedSets.add(s)
                    knownExpansions.add(e)
        # Give indication if an imput did not match anything
        unknownExpansions = options.fan - knownExpansions
        if unknownExpansions:
            logger.warning(
                (f"Unknown fan expansion(s): {', '.join(unknownExpansions)}")
            )

    if options.exclude_expansions:
        # Expand out any wildcards, matching set key or set name in the given language
        expanded_expansions = []
        for e in options.exclude_expansions:
            matches = fnmatch.filter(All_search, e)
            if matches:
                expanded_expansions.extend(matches)
            else:
                expanded_expansions.append(e)

        # Now get the actual sets that are matched above
        options.exclude_expansions = set(
            [e for e in expanded_expansions]
        )  # Remove duplicates
        knownExpansions = set()
        for e in options.exclude_expansions:
            for s in Card.sets:
                if s.lower() == e or Card.sets[s].get("set_name", "").lower() == e:
                    wantedSets.discard(s)
                    knownExpansions.add(e)
        # Give indication if an imput did not match anything
        unknownExpansions = options.exclude_expansions - knownExpansions
        if unknownExpansions:
            logger.warning(
                f"Unknown exclude expansion(s): {', '.join(unknownExpansions)}"
            )

    # Now keep only the cards that are in the sets that have been requested
    keep_cards = []
    for c in cards:
        if c.cardset_tag in wantedSets:
            if options.group_kingdom:
                # Separate non-Kingdom cards (without Randomizer) into new "Extras" set
                if not c.randomizer and Card.sets[c.cardset_tag]["has_extras"]:
                    c.cardset_tag += db.EXPANSION_EXTRA_POSTFIX
            # Add the cardset informaiton to the card and add it to the list of cards to use
            c.cardset = Card.sets[c.cardset_tag].get("set_name", c.cardset_tag)
            keep_cards.append(c)
    cards = keep_cards

    # Now add text to the cards.  Waited as long as possible to catch all groupings
    cards = add_card_text(cards, options.language)

    # Get list of cards from a file
    if options.cardlist:
        cardlist = set()
        cardlist_exclude = set()
        with open(options.cardlist) as cardfile:
            for line in cardfile:
                line = line.strip()
                if not line or line.startswith("#"):
                    pass  # ignore empty and "comment" lines
                elif line.startswith("-"):
                    cardlist_exclude.add(line.lstrip("- \t"))
                else:
                    cardlist.add(line.lstrip("+ \t"))
        if cardlist_exclude:
            cards = [card for card in cards if card.name not in cardlist_exclude]
        if cardlist:
            cards = [card for card in cards if card.name in cardlist]

    # Set up the card sorter
    cardSorter = CardSorter(
        options.order,
        options.language,
        {
            card.card_tag: card.name
            for card in cards
            if "base" in [set_name.lower() for set_name in card.cardset_tags]
        },
    )

    # Optionally remove base cards from expansions that have them
    if not options.base_cards_with_expansion:
        cards = [card for card in cards if not cardSorter.isBaseExpansionCard(card)]

    # Add expansion divider
    if options.expansion_dividers:
        cardnamesByExpansion = defaultdict(dict)
        randomizerCountByExpansion = Counter()
        for c in cards:
            if c.isBlank() or (
                cardSorter.isBaseExpansionCard(c)
                and not options.base_cards_with_expansion
            ):
                continue
            if c.randomizer:
                randomizerCountByExpansion[c.cardset] += 1

            if c.card_tag in cardnamesByExpansion[c.cardset]:
                # Already have one, so just update the count (for extra Curses, Start Decks, etc)
                cardnamesByExpansion[c.cardset][c.card_tag]["count"] += 1
            else:
                # New, so save off information about the card to be used on the expansion divider
                order = 0
                if c.card_tag in cardSorter.baseOrder:
                    # Use the base card ordering
                    order = 100 + cardSorter.baseOrder.index(c.card_tag)
                cardnamesByExpansion[c.cardset][c.card_tag] = {
                    "name": c.name.strip().replace(" ", "&nbsp;"),
                    "randomizer": c.randomizer,
                    "count": 1,
                    "sort": "%03d%s"
                    % (
                        order,
                        cardSorter.get_card_name_sort_key(c.name.strip()),
                    ),
                }

        for set_tag, set_values in Card.sets.items():
            exp = set_values["set_name"]
            if exp in cardnamesByExpansion:
                exp_name = exp

                count = randomizerCountByExpansion[exp]
                Card.sets[set_tag]["count"] = count
                if "no_randomizer" in set_values:
                    if set_values["no_randomizer"]:
                        count = 0

                if not options.expansion_dividers_long_name:
                    if "short_name" in set_values:
                        exp_name = set_values["short_name"]

                card_names = []
                for n in sorted(
                    cardnamesByExpansion[exp].values(), key=lambda x: x["sort"]
                ):
                    if not n["randomizer"]:
                        # Highlight cards without Randomizers
                        n["name"] = "<i>" + n["name"] + "</i>"
                    if n["count"] > 1:
                        # Add number of copies
                        n["name"] = (
                            "{}&nbsp;\u00d7&nbsp;".format(n["count"]) + n["name"]
                        )
                    card_names.append(n["name"])

                c = Card(
                    name=exp_name,
                    cardset=exp,
                    cardset_tag=set_tag,
                    types=("Expansion",),
                    cost=None,
                    description=" | ".join(card_names),
                    extra=set_values.get("set_text", ""),
                    count=count,
                    card_tag=set_tag,
                )
                cards.append(c)

    # Take care of any --only-type-xxx requirements
    if options.only_type_any or options.only_type_all:
        # First make a dictionary for easier lookup of Type name used by the program
        # The index in each case is lower case for easier matching
        # The value in each case is the type index as used in types_en_us.json
        types_lookup = defaultdict(dict)
        for x in Card.type_names:
            types_lookup[x.lower()] = x
            types_lookup[Card.type_names[x].lower()] = x

        # Start the valid lists
        type_unknown = []
        type_known_any = []
        type_known_all = []

        # Assemble a list of valid "any" types.  Options are already lowercase.
        for x in options.only_type_any:
            if x in types_lookup:
                type_known_any.append(types_lookup[x])
            else:
                type_unknown.append(x)

        # Assemble a list of valid "all" types. Options are already lowercase.
        for x in options.only_type_all:
            if x in types_lookup:
                type_known_all.append(types_lookup[x])
            else:
                type_unknown.append(x)

        # Indicate if unknown types are given
        assert not type_unknown, "Error - unknown type(s): {}".format(
            ", ".join(type_unknown)
        )

        # If there are any valid Types left, go through the cards and keep cards that match
        if type_known_any or type_known_all:
            keep_cards = []
            for c in cards:
                if type_known_any and any(x in c.types for x in type_known_any):
                    keep_it = True
                elif type_known_all and all(x in c.types for x in type_known_all):
                    keep_it = True
                else:
                    keep_it = False

                if keep_it:
                    keep_cards.append(c)
            cards = keep_cards

    # Now sort what is left
    cards.sort(key=cardSorter)

    return cards


def calculate_layout(options, cards=None):
    if cards is None:
        cards = []
    # This is in place to allow for test cases to it call directly to get
    options = config_options.clean_opts(options)
    options.dominionCardWidth, options.dominionCardHeight = (
        config_options.parse_cardsize(options.size, options.sleeved)
    )
    options.paperwidth, options.paperheight = config_options.parse_papersize(
        options.papersize
    )
    options.minmarginwidth, options.minmarginheight = config_options.parse_dimensions(
        options.minmargin
    )

    dd = DividerDrawer(options)
    dd.calculatePages(cards)
    return dd


def generate(options):
    cards = db.read_card_data(options)
    assert cards, "No cards after reading"
    cards = filter_sort_cards(cards, options)
    assert cards, "No cards after filtering/sorting"

    dd = calculate_layout(options, cards)

    logger.info(
        f"Paper dimensions: {options.paperwidth / cm:.2f}cm (w) x {options.paperheight / cm:.2f}cm (h)"
    )
    logger.info(
        f"Tab dimensions: {options.dividerWidthReserved / cm:.2f}cm (w) "
        f"x {options.dividerHeightReserved / cm:.2f}cm (h)"
    )
    logger.info(
        f"{options.numDividersHorizontal} dividers horizontally, {options.numDividersVertical} vertically"
    )
    logger.info(
        f"Margins: {options.horizontalMargin / cm:.2f}cm h, {options.verticalMargin / cm:.2f}cm v"
    )

    dd.draw(cards)


def main():
    options = config_options.parse_opts()
    logger.remove()
    logger.add(sys.stderr, level=options.log_level)

    options = config_options.clean_opts(options)
    if options.preview:
        fname = "{}.{}".format(os.path.splitext(options.outfile)[0], "png")
        open(fname, "wb").write(generate_sample(options).getvalue())
    else:
        generate(options)



import copy
import functools
import gzip
import json
import os

import pkg_resources
from loguru import logger

from . import config_options, db
from .cards import Card, CardType

EXPANSION_EXTRA_POSTFIX = " extras"

LANGUAGE_DEFAULT = (
    "en_us"  # the primary language used if a language's parts are missing
)
LANGUAGE_XX = "xx"  # a dummy language for starting translations


@functools.lru_cache()
def get_languages(path="card_db"):
    languages = []
    for name in pkg_resources.resource_listdir("domdiv", path):
        dir_path = os.path.join(path, name)
        if pkg_resources.resource_isdir("domdiv", dir_path):
            cards_file = os.path.join(dir_path, f"cards_{name}.json.gz")
            sets_file = os.path.join(dir_path, f"sets_{name}.json.gz")
            types_file = os.path.join(dir_path, f"types_{name}.json.gz")
            if (
                pkg_resources.resource_exists("domdiv", cards_file)
                and pkg_resources.resource_exists("domdiv", sets_file)
                and pkg_resources.resource_exists("domdiv", types_file)
            ):
                languages.append(name)
    if LANGUAGE_XX in languages:
        languages.remove(LANGUAGE_XX)
    return languages


def get_resource_stream(path):
    return gzip.GzipFile(
        fileobj=pkg_resources.resource_stream("domdiv", path),
    )


@functools.lru_cache()
def get_expansions():
    set_db_filepath = os.path.join("card_db", "sets_db.json.gz")
    with get_resource_stream(set_db_filepath) as setfile:
        set_file = json.loads(setfile.read().decode("utf-8"))
    assert set_file, "Could not load any sets from database"

    fan = []
    official = []
    for s in set_file:
        if EXPANSION_EXTRA_POSTFIX not in s:
            # Make sure these are set either True or False
            set_file[s]["fan"] = set_file[s].get("fan", False)
            if set_file[s]["fan"]:
                fan.append(s)
            else:
                official.append(s)
    fan.sort()
    official.sort()
    return official, fan


@functools.lru_cache()
def get_global_groups():
    type_db_filepath = os.path.join("card_db", "types_db.json.gz")
    with get_resource_stream(type_db_filepath) as typefile:
        type_file = json.loads(typefile.read().decode("utf-8"))
    assert type_file, "Could not load any card types from database"

    group_global_choices = []
    group_global_valid = []
    for t in type_file:
        if "group_global_type" in t:
            group_global_valid.append("-".join(t["card_type"]).lower())
            group_global_choices.append(t["group_global_type"].lower())
    group_global_valid.extend(group_global_choices)
    group_global_valid.sort()
    group_global_choices.sort()
    return group_global_choices, group_global_valid


@functools.lru_cache()
def get_types(language=LANGUAGE_DEFAULT):
    # get a list of valid types
    language = language.lower()
    type_text_filepath = os.path.join("card_db", language, f"types_{language}.json.gz")
    with get_resource_stream(type_text_filepath) as type_text_file:
        type_text = json.loads(type_text_file.read().decode("utf-8"))
    assert type_text, "Could not load type file for %r" % language

    types = [x.lower() for x in type_text]
    types.sort()
    return types


@functools.lru_cache()
def get_label_data():
    labels_db_filepath = os.path.join("card_db", "labels_db.json.gz")
    label_choices = []
    label_keys = []
    label_selections = []
    with get_resource_stream(labels_db_filepath) as labelfile:
        label_info = json.loads(labelfile.read().decode("utf-8"))
    assert label_info, "Could not load label information from database"
    for label in label_info:
        if len(label["names"]) > 0:
            label_keys.append(label["names"][0])
            label_selections.append(
                label["name"] if "name" in label else label["names"][0]
            )
            label_choices.extend(label["names"])
    return label_info, label_keys, label_selections, label_choices


def find_index_of_object(lst=None, attributes=None):
    if lst is None:
        lst = []
    if attributes is None:
        attributes = {}
    # Returns the index of the first object in lst that matches the given attributes.  Otherwise returns None.
    # attributes is a dict of key: value pairs.   Object attributes that are lists are checked to have value in them.
    for i, d in enumerate(lst):
        # Set match to false just in case there are no attributes.
        match = False
        for key, value in attributes.items():
            # if anything does not match, then break out and start the next one.
            match = hasattr(d, key)
            if match:
                test = getattr(d, key, None)
                if isinstance(test, list):
                    match = value in test
                else:
                    match = value == test
            if not match:
                break

        if match:
            # If all the attributes are found, then we have a match
            return i

    # nothing matched
    return None


def read_card_data(options):
    # Read in the card types
    types_db_filepath = os.path.join("card_db", "types_db.json.gz")
    with db.get_resource_stream(types_db_filepath) as typefile:
        Card.types = json.loads(
            typefile.read().decode("utf-8"), object_hook=CardType.decode_json
        )
    assert Card.types, "Could not load any card types from database"

    # extract unique types
    type_list = []
    for c in Card.types:
        type_list = list(set(c.getTypeNames()) | set(type_list))
    # set up the basic type translation.  The actual language will be added later.
    Card.type_names = {}
    for t in type_list:
        Card.type_names[t] = t

    # turn Card.types into a dictionary for later
    Card.types = dict(((c.getTypeNames(), c) for c in Card.types))

    # Read in the card database
    card_db_filepath = os.path.join("card_db", "cards_db.json.gz")
    with get_resource_stream(card_db_filepath) as cardfile:
        cards = json.loads(
            cardfile.read().decode("utf-8"), object_hook=Card.decode_json
        )
    assert cards, "Could not load any cards from database"

    set_db_filepath = os.path.join("card_db", "sets_db.json.gz")
    with get_resource_stream(set_db_filepath) as setfile:
        Card.sets = json.loads(setfile.read().decode("utf-8"))
    assert Card.sets, "Could not load any sets from database"
    new_sets = {}
    for s in Card.sets:
        # Make sure these are set either True or False
        Card.sets[s]["no_randomizer"] = Card.sets[s].get("no_randomizer", False)
        Card.sets[s]["fan"] = Card.sets[s].get("fan", False)
        Card.sets[s]["has_extras"] = Card.sets[s].get("has_extras", True)
        Card.sets[s]["upgrades"] = Card.sets[s].get("upgrades", None)
        new_sets[s] = Card.sets[s]
        # Make an "Extras" set for normal expansions
        if Card.sets[s]["has_extras"]:
            e = s + db.EXPANSION_EXTRA_POSTFIX
            new_sets[e] = copy.deepcopy(Card.sets[s])
            new_sets[e]["set_name"] = "*" + s + db.EXPANSION_EXTRA_POSTFIX + "*"
            new_sets[e]["no_randomizer"] = True
            new_sets[e]["has_extras"] = False
    Card.sets = new_sets

    # Remove the Trash card. Do early before propagating to various sets.
    if options.no_trash:
        i = find_index_of_object(cards, {"card_tag": "Trash"})
        if i is not None:
            del cards[i]

    # Repackage Curse cards into 10 per divider. Do early before propagating to various sets.
    if options.curse10:
        i = find_index_of_object(cards, {"card_tag": "Curse"})
        if i is not None:
            new_cards = []
            cards_remaining = cards[i].getCardCount()
            while cards_remaining > 10:
                # make a new copy of the card and set count to 10
                new_card = copy.deepcopy(cards[i])
                new_card.setCardCount(10)
                new_cards.append(new_card)
                cards_remaining -= 10

            # Adjust original Curse card to the remaining cards (should be 10)
            cards[i].setCardCount(cards_remaining)
            # Add the new dividers
            cards.extend(new_cards)

    # Add any blank cards
    if options.include_blanks > 0:
        for _ in range(0, options.include_blanks):
            c = Card(
                card_tag="Blank",
                cardset=config_options.EXPANSION_GLOBAL_GROUP,
                cardset_tag=config_options.EXPANSION_GLOBAL_GROUP,
                cardset_tags=[config_options.EXPANSION_GLOBAL_GROUP],
                randomizer=False,
                types=("Blank",),
            )
            cards.append(c)

    # Create Start Deck dividers. 4 sets. Adjust totals for other cards, too.
    # Do early before propagating to various sets.
    # The card database contains one prototype divider that needs to be either duplicated or deleted.
    if options.start_decks:
        # Find the index to the individual cards that need changed in the cards list
        StartDeck_index = find_index_of_object(cards, {"card_tag": "Start Deck"})
        Copper_index = find_index_of_object(cards, {"card_tag": "Copper"})
        Estate_index = find_index_of_object(cards, {"card_tag": "Estate"})
        if Copper_index is None or Estate_index is None or StartDeck_index is None:
            # Something is wrong, can't find one or more of the cards that need to change
            logger.warning("Cannot create Start Decks")

            # Remove the Start Deck prototype if we can
            if StartDeck_index is not None:
                del cards[StartDeck_index]
        else:
            # Start Deck Constants
            STARTDECK_COPPERS = 7
            STARTDECK_ESTATES = 3
            STARTDECK_NUMBER = 4

            # Add correct card counts to Start Deck prototype.  This will be used to make copies.
            cards[StartDeck_index].setCardCount(STARTDECK_COPPERS)
            cards[StartDeck_index].addCardCount([int(STARTDECK_ESTATES)])

            # Make new Start Deck Dividers and adjust the corresponding card counts
            for x in range(0, STARTDECK_NUMBER):
                # Add extra copies of the Start Deck prototype.
                # But don't need to add the first one again, since the prototype is already there.
                if x > 0:
                    cards.append(copy.deepcopy(cards[StartDeck_index]))
                    # Note: By appending, it should not change any of the index values being used

                # Remove Copper and Estate card counts from their dividers
                cards[Copper_index].setCardCount(
                    cards[Copper_index].getCardCount() - STARTDECK_COPPERS
                )
                cards[Estate_index].setCardCount(
                    cards[Estate_index].getCardCount() - STARTDECK_ESTATES
                )
    else:
        # Remove Start Deck prototype.  It is not needed.
        StartDeck_index = find_index_of_object(cards, {"card_tag": "Start Deck"})
        if StartDeck_index is not None:
            del cards[StartDeck_index]

    # Set cardset_tag and expand cards that are used in multiple sets
    new_cards = []
    for card in cards:
        sets = list(card.cardset_tags)
        if len(sets) > 0:
            # Set and save the first one
            card.cardset_tag = sets.pop(0)
            new_cards.append(card)
            for s in sets:
                # for the rest, create a copy of the first
                if s:
                    new_card = copy.deepcopy(card)
                    new_card.cardset_tag = s
                    new_cards.append(new_card)
    cards = new_cards

    # Make sure each card has the right image file.
    for card in cards:
        card.image = card.setImage()

    return cards



import os
from zipfile import ZIP_DEFLATED, ZipFile

import domdiv
import domdiv.config_options
import domdiv.main

gen_dir = "sumpfork_dividers"
prefix = f"{gen_dir}/sumpfork_dominion_tabs_"
postfix = "v" + domdiv.__version__ + ".pdf"
argsets = [
    ("", ""),
    ("--orientation=vertical", "vertical_"),
    ("--papersize=A4", "A4_"),
    ("--papersize=A4 --orientation=vertical", "vertical_A4_"),
    ("--size=sleeved", "sleeved_"),
    ("--size=sleeved --orientation=vertical", "vertical_sleeved_"),
]
additional = ["--expansion-dividers", "--tab-artwork-resolution=300"]


def run_generator(args, main):
    args = f"{args} --font-dir local_fonts --outfile {prefix}{main}{postfix}"
    args = args.split()
    fname = args[-1]
    print(args)
    print(":::Generating " + fname)
    options = domdiv.config_options.parse_opts(args)
    options = domdiv.config_options.clean_opts(options)
    domdiv.main.generate(options)
    return fname


def make_bgg_release():
    if not os.path.exists(gen_dir):
        print(f"Making dir '{gen_dir}'")
        os.mkdir(gen_dir)

    fnames = [
        run_generator(args[0] + " " + " ".join(additional), args[1]) for args in argsets
    ]
    print(fnames)

    with ZipFile(
        f"{gen_dir}/sumpfork_dominion_tabs_v" + domdiv.__version__ + ".zip",
        "w",
        ZIP_DEFLATED,
        compresslevel=9,
    ) as zip:
        for f in fnames:
            zip.write(f)



from __future__ import print_function

import pytest

from domdiv import config_options, db, main


def get_clean_opts(opts):
    options = config_options.parse_opts(opts)
    options = config_options.clean_opts(options)
    return options


def test_standard_opts():
    # should be the default
    options = get_clean_opts([])
    main.generate(options)


@pytest.mark.parametrize("lang", db.get_languages("card_db"))
def test_languages(lang):
    print("checking " + lang)
    options = get_clean_opts(["--language={}".format(lang)])
    main.generate(options)


def test_grouped():
    options = get_clean_opts(["--special-card-groups"])
    main.generate(options)


def test_resolution():
    options = get_clean_opts(["--tab-artwork-resolution=300"])
    main.generate(options)



import codecs
import gzip
import json
import os

LANGUAGE_DEFAULT = "en_us"  # default language, which takes priority
LANGUAGE_XX = "xx"  # language for starting a translation


# Multikey sort
# see: http://stackoverflow.com/questions/1143671/python-sorting-list-of-dictionaries-by-multiple-keys
def multikeysort(items, columns):
    from operator import itemgetter

    for c in columns[::-1]:
        items = sorted(items, key=itemgetter(c))
    return items


def get_languages(card_db_dir):
    languages = get_lang_dirs(card_db_dir)
    languages.remove(LANGUAGE_DEFAULT)
    languages.insert(0, LANGUAGE_DEFAULT)
    if LANGUAGE_XX not in languages:
        languages.append(LANGUAGE_XX)
    return languages


def get_lang_dirs(path):
    # Find all valid languages.
    lang_dirs = []
    for name in os.listdir(path):
        dir_path = os.path.join(path, name)
        if os.path.isdir(dir_path):
            cards_file = os.path.join(dir_path, "cards_" + name + ".json")
            sets_file = os.path.join(dir_path, "sets_" + name + ".json")
            if os.path.isfile(cards_file) and os.path.isfile(sets_file):
                lang_dirs.append(name)
    return lang_dirs


def get_json_data(json_file_path):
    print(("reading {}".format(json_file_path)))
    # Read in the json from the specified file
    with codecs.open(json_file_path, "r", "utf-8") as json_file:
        data = json.load(json_file)
    assert data, "Could not load json at: '%r' " % json_file_path
    return data


def load_language_cards(lang, card_db_dir):
    #  contruct the cards json file name
    lang_file = f"cards_{lang}.json"
    fname = os.path.join(card_db_dir, lang, lang_file)
    if os.path.isfile(fname):
        lang_data = get_json_data(fname)
    else:
        lang_data = {}
    return lang_data


def check_compressed_json_change(gzip_fname, data):
    if os.path.exists(gzip_fname):
        with gzip.open(gzip_fname, "rt", encoding="utf-8") as f:
            existing_contents = f.read()
            new_contents = json.dumps(data, indent=4, ensure_ascii=False) + "\n"
            return existing_contents == new_contents


def write_data(data, fname, do_gzip=True, check_changed=True):
    #  Process the file
    final_name = f"{fname}.gz" if do_gzip else fname
    mode = "wt" if do_gzip else "w"
    print(f"writing {final_name}")

    open_func = gzip.open if do_gzip else open

    if do_gzip and check_changed:
        if os.path.exists(final_name):
            if check_compressed_json_change(final_name, data):
                print(f"not rewriting {fname} - contents didn't change")
                return

    with open_func(final_name, mode, encoding="utf-8") as lang_out:
        json.dump(data, lang_out, indent=4, ensure_ascii=False)
        lang_out.write("\n")


def write_language_cards(cards, lang, card_db_dir):
    lang_file = f"cards_{lang}.json"
    fname = os.path.join(card_db_dir, lang, lang_file)
    write_data(cards, fname)


def load_card_data(card_db_dir):
    ###########################################################################
    #  Get the cards_db information
    #  Store in a list in the order found in cards[]. Ordered as follows:
    #  1. card_tags, 2. group_tags, 3. super groups
    ###########################################################################

    # Get the card data
    card_data = get_json_data(os.path.join(card_db_dir, "cards_db.json"))

    # Sort the cardset_tags
    for card in card_data:
        card["cardset_tags"].sort()
        # But put all the base cards together by moving to front of the list
        if "base" in card["cardset_tags"]:
            card["cardset_tags"].remove("base")
            card["cardset_tags"].insert(0, "base")

    # Sort the cards by cardset_tags, then card_tag
    return multikeysort(card_data, ["cardset_tags", "card_tag"])



import os
import sys

import configargparse
import reportlab.lib.pagesizes as pagesizes
from loguru import logger
from reportlab.lib.units import cm

from . import db

LOCATION_CHOICES = ["tab", "body-top", "hide"]
NAME_ALIGN_CHOICES = ["left", "right", "centre", "edge"]
TAB_SIDE_CHOICES = [
    "left",
    "right",
    "left-alternate",
    "right-alternate",
    "left-flip",
    "right-flip",
    "centre",
    "full",
]
TEXT_CHOICES = ["card", "rules", "blank"]
LINE_CHOICES = ["line", "dot", "cropmarks", "line-cropmarks", "dot-cropmarks"]

HEAD_CHOICES = ["tab", "strap", "cover", "none"]
TAIL_CHOICES = ["tab", "strap", "cover", "folder", "none"]
FACE_CHOICES = ["front", "back"]
SPINE_CHOICES = ["name", "types", "tab", "blank"]

EDITION_CHOICES = ["1", "2", "latest", "upgrade", "removed", "all"]

ORDER_CHOICES = ["expansion", "global", "colour", "cost"]

EXPANSION_GLOBAL_GROUP = "extras"


def add_opt(options, option, value):
    assert not hasattr(options, option)
    setattr(options, option, value)


def parse_opts(cmdline_args=None):
    parser = configargparse.ArgParser(
        formatter_class=configargparse.ArgumentDefaultsHelpFormatter,
        description="Generate Dominion Dividers",
        epilog="Source can be found at 'https://github.com/sumpfork/dominiontabs'. "
        "An online version can be found at 'http://domdiv.bgtools.net/'. ",
    )

    # Basic Divider Information
    group_basic = parser.add_argument_group(
        "Basic Divider Options", "Basic choices for the dividers."
    )
    group_basic.add_argument(
        "--outfile",
        "-o",
        dest="outfile",
        default="dominion_dividers.pdf",
        help="The output file name.",
    )
    group_basic.add_argument(
        "--papersize",
        dest="papersize",
        default=None,
        help="The size of paper to use; '<%%f>x<%%f>' (size in cm), or 'A4', or 'LETTER'. "
        "If not specified, it will default to system defaults, and if the system defaults "
        "are not found, then to 'LETTER'.",
    )
    group_basic.add_argument(
        "--language",
        "-l",
        dest="language",
        default=db.LANGUAGE_DEFAULT,
        choices=db.get_languages(),
        help="Language of divider text.",
    )
    group_basic.add_argument(
        "--font-dir",
        help="A directory path to scan for font files, preferring them over fonts in the domdiv package",
    )
    group_basic.add_argument(
        "--orientation",
        choices=["horizontal", "vertical"],
        dest="orientation",
        default="horizontal",
        help="Either horizontal or vertical divider orientation.",
    )
    group_basic.add_argument(
        "--size",
        dest="size",
        default="normal",
        help="Dimensions of the cards to use with the dividers '<%%f>x<%%f>' (size in cm), "
        "or 'normal' = '9.1x5.9', or 'sleeved' = '9.4x6.15'.",
    )
    group_basic.add_argument(
        "--sleeved", action="store_true", dest="sleeved", help="Same as --size=sleeved."
    )
    group_basic.add_argument(
        "--order",
        choices=ORDER_CHOICES,
        default="expansion",
        dest="order",
        help="Sort order for the dividers: "
        " 'global' will sort by card name;"
        " 'expansion' will sort by expansion, then card name;"
        " 'colour' will sort by card type, then card name;"
        " 'cost' will sort by expansion, then card cost, then name.",
    )

    # Divider Body
    group_body = parser.add_argument_group(
        "Divider Body", "Changes what is displayed on the body of the dividers."
    )
    group_body.add_argument(
        "--front",
        choices=TEXT_CHOICES,
        dest="text_front",
        default="card",
        help="Text to print on the front of the divider; "
        "'card' will print the text from the game card; "
        "'rules' will print additional rules for the game card; "
        "'blank' will not print text on the divider.",
    )
    group_body.add_argument(
        "--back",
        choices=TEXT_CHOICES + ["none"],
        dest="text_back",
        default="rules",
        help="Text to print on the back of the divider; "
        "'card' will print the text from the game card; "
        "'rules' will print additional rules for the game card; "
        "'blank' will not print text on the divider; "
        "'none' will prevent the back pages from printing. ",
    )
    group_body.add_argument(
        "--count",
        action="store_true",
        dest="count",
        help="Display the card count on the body of card dividers "
        "and the randomizer count on the body of expansion dividers.",
    )
    group_body.add_argument(
        "--types",
        action="store_true",
        dest="types",
        help="Display card type on the body of the divider.",
    )

    # Divider Tab
    group_tab = parser.add_argument_group(
        "Divider Tab", "Changes what is displayed on on the Divider Tab."
    )
    group_tab.add_argument(
        "--tab-side",
        choices=TAB_SIDE_CHOICES,
        dest="tab_side",
        default="right-alternate",
        help="Alignment of tab; "
        "'left'/'right'/'centre' sets the starting side of the tabs; "
        "'full' will force all label tabs to be full width of the divider; sets --tab_number 1 "
        "'left-alternate' will start on the left and then toggle between left and right for the tabs,"
        " sets --tab_number 2; "
        "'right-alternate' will start on the right and then toggle between right and left for the tabs,"
        " sets --tab_number 2; "
        "'left-flip' like left-alternate, but the right will be flipped front/back with tab on left,"
        " sets --tab_number 2; "
        "'right-flip' like right-alternate, but the left will be flipped front/back with tab on right,"
        " sets --tab_number 2; ",
    )
    group_tab.add_argument(
        "--tab-number",
        type=int,
        default=2,
        help="The number of tabs. When set to 1, all tabs are on the same side (specified by --tab_side). "
        "When set to 2, tabs will alternate between left and right. (starting side specified by --tab_side). "
        "When set > 2, the first tab will be on left/right side specified by --tab_side, then the rest "
        "of the tabs will be evenly spaced until ending on the opposite side. Then the cycle repeats. "
        "May be overriden by some options of --tab_side.",
    )
    group_tab.add_argument(
        "--tab-serpentine",
        action="store_true",
        help="Affects the order of tabs.  When not selected, tabs will progress from the starting side (left/right) "
        "to the opposite side (right/left), and then repeat (e.g., left to right, left to right, etc.). "
        "When selected, the order is changed to smoothly alternate between the two sides "
        "(e.g., left to right, to left, to right, etc.) "
        "Only valid if --tab_number > 2.",
    )
    group_tab.add_argument(
        "--tab-name-align",
        choices=NAME_ALIGN_CHOICES + ["center"],
        dest="tab_name_align",
        default="left",
        help="Alignment of text on the tab; "
        "The 'edge' option will align the card name to the outside edge of the "
        "tab, so that when using tabs on alternating sides, "
        "the name is less likely to be hidden by the tab in front "
        "(edge will revert to left when tab_side is full since there is no edge in that case).",
    )
    group_tab.add_argument(
        "--tabwidth",
        type=float,
        default=4.0,
        help="Width in cm of stick-up tab (ignored if --tab_side is 'full' or --tabs_only is used).",
    )
    group_tab.add_argument(
        "--cost",
        action="append",
        choices=LOCATION_CHOICES,
        help="Where to display the card cost; may be set to "
        "'hide' to indicate it should not be displayed, or "
        "given multiple times to show it in multiple places. "
        "(If not given, will default to 'tab'.)",
    )
    group_tab.add_argument(
        "--set-icon",
        action="append",
        choices=LOCATION_CHOICES,
        help="Where to display the set icon; may be set to "
        "'hide' to indicate it should not be displayed, or "
        "given multiple times to show it in multiple places. "
        "(If not given, will default to 'tab'.)",
    )
    group_tab.add_argument(
        "--no-tab-artwork",
        action="store_true",
        dest="no_tab_artwork",
        help="Don't show background artwork on tabs.",
    )
    group_tab.add_argument(
        "--tab-artwork-opacity",
        type=float,
        default=1.0,
        help="Multiply opacity of tab background art by this value; "
        "can be used to make text show up clearer on dark backrounds, "
        "particularly on printers that output darker than average",
    )
    group_tab.add_argument(
        "--tab-artwork-resolution",
        type=int,
        default=0,
        help="Limit the DPI resolution of tab background art.  "
        "If nonzero, any higher-resolution images will be resized to "
        "reduce output file size.",
    )
    group_tab.add_argument(
        "--use-text-set-icon",
        action="store_true",
        dest="use_text_set_icon",
        help="Use text/letters to represent a card's set instead of the set icon.",
    )
    group_tab.add_argument(
        "--use-set-icon",
        action="store_true",
        dest="use_set_icon",
        help="Use set icon instead of a card icon.  Applies to Promo cards.",
    )
    group_tab.add_argument(
        "--expansion-reset-tabs",
        action="store_true",
        dest="expansion_reset_tabs",
        help="When set, the tabs are restarted (left/right) at the beginning of each expansion. "
        "If not set, the tab pattern will continue from one expansion to the next. ",
    )

    # Expanion Dividers
    group_expansion = parser.add_argument_group(
        "Expansion Dividers", "Adding separator dividers for each expansion."
    )
    group_expansion.add_argument(
        "--expansion-dividers",
        action="store_true",
        dest="expansion_dividers",
        help="Add dividers describing each expansion set. "
        "A list of cards in the expansion will be shown on the front of the divider.",
    )
    group_expansion.add_argument(
        "--centre-expansion-dividers",
        action="store_true",
        dest="centre_expansion_dividers",
        help="Centre the tabs on expansion dividers (same width as dividers.)",
    )
    group_expansion.add_argument(
        "--full-expansion-dividers",
        action="store_true",
        dest="full_expansion_dividers",
        help="Full width expansion dividers.",
    )
    group_expansion.add_argument(
        "--expansion-dividers-long-name",
        action="store_true",
        dest="expansion_dividers_long_name",
        help="Use the long name with edition information on the expansion divider tab. "
        "Without this, the shorter expansion name is used on the expansion divider tab.",
    )

    # Divider Selection
    group_select = parser.add_argument_group(
        "Divider Selection", "What expansions are used, and grouping of dividers."
    )
    group_select.add_argument(
        "--expansions",
        "--expansion",
        nargs="*",
        action="append",
        dest="expansions",
        help="Limit dividers to only the specified expansions. "
        "If no limits are set, then the latest expansions are included. "
        "Expansion names can also be given in the language specified by "
        "the --language parameter. Any expansion with a space in the name must "
        "be enclosed in double quotes. This may be called multiple times. "
        "Values are not case sensitive. Wildcards may be used: "
        "'*' any number of characters, '?' matches any single character, "
        "'[seq]' matches any character in seq, and '[!seq]' matches any character not in seq. "
        "For example, 'dominion*' will match all expansions that start with 'dominion'. "
        "Choices available in all languages include: {}".format(
            ", ".join("%s" % x for x in db.get_expansions()[0])
        ),
    )
    group_select.add_argument(
        "--fan",
        nargs="*",
        action="append",
        dest="fan",
        help="Add dividers from the specified fan made expansions. "
        "If this option is not used, no fan expansions will be included. "
        "Fan made expansion names can also be given in the language specified by "
        "the --language parameter. Any fan expansion with a space in the name must "
        "be enclosed in double quotes. This may be called multiple times. "
        "Values are not case sensitive. Wildcards may be used: "
        "'*' any number of characters, '?' matches any single character, "
        "'[seq]' matches any character in seq, and '[!seq]' matches any character not in seq. "
        "Choices available in all languages include: {}".format(
            ", ".join("%s" % x for x in db.get_expansions()[1])
        ),
    )
    group_select.add_argument(
        "--exclude-expansions",
        "--exclude-expansion",
        nargs="*",
        action="append",
        metavar="EXCLUDED",
        dest="exclude_expansions",
        help="Limit dividers to not include the specified expansions. "
        "Useful if you want all the expansions, except for one or two. "
        "If an expansion is explicitly specified with both '--expansion' and "
        "'--exclude-expansion', then '--exclude-expansion' wins, and the "
        "expansion is NOT included. Expansion names can also be given in the "
        "language specified by the --language parameter. Any expansion with a "
        "space in the name must be enclosed in double quotes. This may be "
        "called multiple times.  Values are not case sensitive. Wildcards may "
        "be used. See the help for '--expansion' for details on wildcards. May "
        "be the name of an official expansion or fan expansion - see the help "
        "for --expansion and --fan for a list of possible names.",
    )
    group_select.add_argument(
        "--edition",
        choices=EDITION_CHOICES,
        dest="edition",
        help="Editions to include: "
        "'1' is for all 1st Editions; "
        "'2' is for all 2nd Editions; "
        "'upgrade' is for all upgrade cards for each expansion; "
        "'removed' is for all removed cards for each expansion; "
        "'latest' is for the latest edition for each expansion; "
        "'all' is for all editions of expansions, upgrade cards, and removed cards; "
        " This can be combined with other options to refine the expansions to include in the output."
        " (default: all)",
    )
    group_select.add_argument(
        "--upgrade-with-expansion",
        action="store_true",
        dest="upgrade_with_expansion",
        help="Include any new edition upgrade cards with the expansion being upgraded.",
    )
    group_select.add_argument(
        "--base-cards-with-expansion",
        action="store_true",
        help="Print the base cards as part of the expansion (i.e., a divider for 'Silver' "
        "will be printed as both a 'Dominion' card and as an 'Intrigue 1st Edition' card). "
        "If this option is not given, all base cards are placed in their own 'Base' expansion.",
    )
    group_select.add_argument(
        "--group-special",
        "--special-card-groups",
        action="store_true",
        dest="group_special",
        help="Group cards that generally are used together "
        "(e.g., Shelters, Tournament and Prizes, Urchin/Mercenary, etc.).",
    )
    group_select.add_argument(
        "--group-kingdom",
        action="store_true",
        dest="group_kingdom",
        help="Group cards that have randomizers into the expansion, "
        "and those that don't have randomizers into the expansion's 'Extra' section.",
    )
    group_select.add_argument(
        "--group-global",
        nargs="*",
        action="append",
        dest="group_global",
        help="Group all cards of the specified types across all expansions into one 'Extras' divider. "
        "This may be called multiple times. Values are not case sensitive. "
        "Choices available include: {}".format(
            ", ".join("%s" % x for x in db.get_global_groups()[1])
        ),
    )
    group_select.add_argument(
        "--no-trash",
        action="store_true",
        dest="no_trash",
        help="Exclude Trash from cards.",
    )
    group_select.add_argument(
        "--curse10",
        action="store_true",
        dest="curse10",
        help="Package Curse cards into groups of ten cards.",
    )
    group_select.add_argument(
        "--start-decks",
        action="store_true",
        dest="start_decks",
        help="Include four start decks with the Base cards.",
    )
    group_select.add_argument(
        "--include-blanks",
        type=int,
        default=0,
        help="Number of blank dividers to include.",
    )
    group_select.add_argument(
        "--exclude-events",
        action="store_true",
        help="Group all 'Event' cards across all expansions into one divider."
        "Same as '--group-global Events'",
    )
    group_select.add_argument(
        "--exclude-landmarks",
        action="store_true",
        help="Group all 'Landmark' cards across all expansions into one divider."
        "Same as '--group-global landmarks'",
    )
    group_select.add_argument(
        "--exclude-projects",
        action="store_true",
        help="Group all 'Project' cards across all expansions into one divider."
        "Same as '--group-global projects'",
    )
    group_select.add_argument(
        "--exclude-ways",
        action="store_true",
        help="Group all 'Way' cards across all expansions into one divider."
        "Same as '--group-global ways'",
    )
    group_select.add_argument(
        "--exclude-traits",
        action="store_true",
        help="Group all 'Trait' cards across all expansions into one divider."
        "Same as '--group-global traits'",
    )
    group_select.add_argument(
        "--only-type-any",
        "--only-type",
        "--type-any",
        nargs="*",
        action="append",
        dest="only_type_any",
        help="Limit dividers to only those with the specified types. "
        "A divider is kept if ANY of the provided types are associated with the divider. "
        "Default is all types are included. "
        "Any type with a space in the name must be enclosed in double quotes. "
        "Values are not case sensitive. "
        "Choices available in all languages include: {}".format(
            ", ".join("%s" % x for x in db.get_types())
        ),
    )
    group_select.add_argument(
        "--only-type-all",
        "--type-all",
        nargs="*",
        action="append",
        dest="only_type_all",
        help="Limit dividers to only those with the specified types. "
        "A divider is kept if ALL of the provided types are associated with the divider. "
        "Any type with a space in the name must be enclosed in double quotes. "
        "Values are not case sensitive. "
        "Choices available in all languages include: {}".format(
            ", ".join("%s" % x for x in db.get_types())
        ),
    )

    # Divider Sleeves/Wrappers
    group_wrapper = parser.add_argument_group(
        "Card Sleeves/Wrappers", "Generating dividers that are card sleeves/wrappers."
    )
    group_wrapper.add_argument(
        "--wrapper",
        action="store_true",
        dest="wrapper_meta",
        help="Draw sleeves (wrappers) instead of dividers for the cards. "
        "Same as --head=strap --tail=folder",
    )
    group_wrapper.add_argument(
        "--pull-tab",
        action="store_true",
        dest="pull_tab_meta",
        help="Draw folding pull tabs instead of dividers for the cards. "
        "Same as --head=tab --tail=cover",
    )
    group_wrapper.add_argument(
        "--tent",
        action="store_true",
        dest="tent_meta",
        help="Draw folding tent covers instead of dividers for the cards. "
        "Same as --head=cover --head-facing=back --head-text=back "
        "--tail=tab --tail-facing=front",
    )
    group_wrapper.add_argument(
        "--head",
        choices=HEAD_CHOICES,
        dest="head",
        default="tab",
        help="Top tab or wrapper type: "
        "'tab' for divider tabs, "
        "'strap' for longer folding tabs, "
        "'cover' for matchbook-style folding covers, "
        "or 'none' to leave the top edge plain. "
        "The folding options create a top spine that you can customize "
        "with --spine.",
    )
    group_wrapper.add_argument(
        "--tail",
        choices=TAIL_CHOICES,
        dest="tail",
        default="none",
        help="Bottom tab or wrapper type: "
        "'tab' for a bottom tab banner, "
        "'strap' for a pull tab under the cards, "
        "'cover' for a simple back cover, "
        "'folder' to create tab folders, "
        "or 'none' to leave the bottom edge plain.",
    )
    group_wrapper.add_argument(
        "--head-facing",
        choices=FACE_CHOICES,
        dest="head_facing",
        default="front",
        help="Text orientation for top tabs and wrappers: "
        "'front' shows the text upright when flat, "
        "'back' shows it upright when folded over.",
    )
    group_wrapper.add_argument(
        "--tail-facing",
        choices=FACE_CHOICES,
        dest="tail_facing",
        default="back",
        help="Text orientation for tail wrappers: "
        "'front' shows the text upright when flat, "
        "'back' shows it upright when folded under.",
    )
    group_wrapper.add_argument(
        "--head-text",
        choices=TEXT_CHOICES + FACE_CHOICES,
        dest="head_text",
        default="blank",
        help="Text to print on top cover panels: "
        "'card' shows the text from the game card, "
        "'rules' shows additional rules for the game card, "
        "'blank' leaves the panel blank; "
        "'front' uses the same setting as --front; "
        "'back' uses the same setting as --back.",
    )
    group_wrapper.add_argument(
        "--tail-text",
        choices=TEXT_CHOICES + FACE_CHOICES,
        dest="tail_text",
        default="back",
        help="Text to print on bottom folder panels: "
        "'card' shows the text from the game card, "
        "'rules' shows additional rules for the game card, "
        "'blank' leaves the panel blank; "
        "'front' uses the same setting as --front; "
        "'back' uses the same setting as --back.",
    )
    group_wrapper.add_argument(
        "--head-height",
        type=float,
        default=0.0,
        help="Height of the top panel in centimeters "
        "(a value of 0 uses tab height or card height as appropriate).",
    )
    group_wrapper.add_argument(
        "--tail-height",
        type=float,
        default=0.0,
        help="Height of the bottom panel in centimeters "
        "(a value of 0 uses tab height or card height as appropriate).",
    )
    group_wrapper.add_argument(
        "--spine",
        choices=SPINE_CHOICES,
        dest="spine",
        default="name",
        help="Text to print on the spine of top covers: "
        "'name' prints the card name; "
        "'type' prints the card type; "
        "'tab' prints tab text and graphics; "
        "'blank' leaves the spine blank. "
        "This is only valid with folding --head options.",
    )
    group_wrapper.add_argument(
        "--thickness",
        type=float,
        default=2.0,
        help="Thickness of a stack of 60 cards (Copper) in centimeters. "
        "Typically unsleeved cards are 2.0, thin sleeved cards are 2.4, and thick sleeved cards are 3.2. "
        "This is only valid with --wrapper or other folding options.",
    )
    group_wrapper.add_argument(
        "--sleeved-thick",
        action="store_true",
        dest="sleeved_thick",
        help="Same as --size=sleeved --thickness 3.2.",
    )
    group_wrapper.add_argument(
        "--sleeved-thin",
        action="store_true",
        dest="sleeved_thin",
        help="Same as --size=sleeved --thickness 2.4.",
    )
    group_wrapper.add_argument(
        "--notch",
        action="store_true",
        dest="notch",
        help="Creates thumb notches opposite to the divider tabs, "
        "which can make it easier to remove cards from wrappers or stacks. "
        "Equivalent to --notch-length=1.5 --notch-height=0.25",
    )
    group_wrapper.add_argument(
        "--notch-length",
        type=float,
        default=0.0,
        help="Sets the length of thumb notches in centimeters.",
    )
    group_wrapper.add_argument(
        "--notch-height",
        type=float,
        default=0.0,
        help="Sets the height of thumb notches in centimeters.",
    )

    # Printing
    group_printing = parser.add_argument_group(
        "Printing", "Changes how the Dividers are printed."
    )
    group_printing.add_argument(
        "--minmargin",
        dest="minmargin",
        default="1x1",
        help="Page margin in cm in the form '<%%f>x<%%f>', left/right x top/bottom).",
    )
    group_printing.add_argument(
        "--cropmarks",
        action="store_true",
        dest="cropmarks",
        help="Print crop marks on both sides, rather than tab outlines on the front side.",
    )
    group_printing.add_argument(
        "--linewidth",
        type=float,
        default=0.1,
        help="Width of lines for card outlines and crop marks.",
    )
    group_printing.add_argument(
        "--front-offset",
        type=float,
        dest="front_offset",
        default=0,
        help="Front page horizontal offset points to shift to the right. Only needed for some printers.",
    )
    group_printing.add_argument(
        "--front-offset-height",
        type=float,
        dest="front_offset_height",
        default=0,
        help="Front page vertical offset points to shift upward. Only needed for some printers.",
    )
    group_printing.add_argument(
        "--back-offset",
        type=float,
        dest="back_offset",
        default=0,
        help="Back page horizontal offset points to shift to the right. Only needed for some printers.",
    )
    group_printing.add_argument(
        "--back-offset-height",
        type=float,
        dest="back_offset_height",
        default=0,
        help="Back page vertical offset points to shift upward. Only needed for some printers.",
    )
    group_printing.add_argument(
        "--vertical-gap",
        type=float,
        default=0.0,
        help="Vertical gap between dividers in centimeters.",
    )
    group_printing.add_argument(
        "--horizontal-gap",
        type=float,
        default=0.0,
        help="Horizontal gap between dividers in centimeters.",
    )
    group_printing.add_argument(
        "--no-page-footer",
        action="store_true",
        dest="no_page_footer",
        help="Do not print the expansion name at the bottom of the page.",
    )
    group_printing.add_argument(
        "--num-pages",
        type=int,
        default=-1,
        help="Stop generating dividers after this many pages, -1 for all.",
    )
    group_printing.add_argument(
        "--tabs-only",
        action="store_true",
        dest="tabs_only",
        help="Draw only the divider tabs and no divider outlines. "
        "Used to print the divider tabs on labels.",
    )
    group_printing.add_argument(
        "--black-tabs",
        action="store_true",
        help="In tabs-only mode, draw tabs on black background",
    )
    group_printing.add_argument(
        "--linetype",
        choices=LINE_CHOICES,
        dest="linetype",
        default="line",
        help="The divider outline type. "
        "'line' will print a solid line outlining the divider; "
        "'dot' will print a dot at each corner of the divider; "
        "'cropmarks' will print cropmarks for the divider; "
        "'line-cropmarks' will combine 'line' and 'cropmarks'; "
        "'dot-cropmarks' will combine 'dot' and 'cropmarks'",
    )
    group_printing.add_argument(
        "--cropmarkLength",
        type=float,
        default=0.2,
        help="Length of actual drawn cropmark in centimeters.",
    )
    group_printing.add_argument(
        "--cropmarkSpacing",
        type=float,
        default=0.1,
        help="Spacing between card and the start of the cropmark in centimeters.",
    )
    group_printing.add_argument(
        "--rotate",
        type=int,
        choices=[0, 90, 180, 270],
        default=0,
        help="Divider degrees of rotation relative to the page edge. "
        "No optimization will be done on the number of dividers per page.",
    )
    group_printing.add_argument(
        "--label",
        dest="label_name",
        choices=db.get_label_data()[3],
        default=None,
        help="Use preset label dimentions. Specify a label name. "
        "This will override settings that conflict with the preset label settings.",
    )
    group_printing.add_argument(
        "--info",
        action="store_true",
        dest="info",
        help="Add a page that has all the options used for the file.",
    )
    group_printing.add_argument(
        "--info-all",
        action="store_true",
        dest="info_all",
        help="Same as --info, but includes pages with all the possible options that can be used.",
    )
    group_printing.add_argument(
        "--preview",
        action="store_true",
        help="Only generate a preview png image of the first page",
    )
    group_printing.add_argument(
        "--preview-resolution",
        type=int,
        default=150,
        help="resolution in DPI to render preview at, for --preview option",
    )
    # Special processing
    group_special = parser.add_argument_group(
        "Miscellaneous", "These options are generally not used."
    )
    group_special.add_argument(
        "--cardlist",
        dest="cardlist",
        help="Path to file that enumerates each card on its own line to be included or excluded."
        " To include a card, add its card name on a line.  The name can optionally be preceeded by '+'."
        " To exclude a card, add its card name on a line preseeded by a '-'"
        " If any card is included by this method, only cards specified in this file will be printed.",
    )
    group_special.add_argument(
        "-c",
        is_config_file=True,
        help="Use the specified configuration file to provide options. "
        "Command line options override options from this file.",
    )
    group_special.add_argument(
        "-w",
        is_write_out_config_file_arg=True,
        help="Write out the given options to the specified configuration file.",
    )
    group_special.add_argument(
        "--log-level",
        default="WARNING",
        help="Set the logging level.",
        choices=["TRACE", "DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
    )
    options = parser.parse_args(args=cmdline_args)
    # Need to do these while we have access to the parser
    options.argv = sys.argv if options.info or options.info_all else None
    options.help = parser.format_help() if options.info_all else None
    return options


def clean_opts(options):
    if "center" in options.tab_side:
        options.tab_side = str(options.tab_side).replace("center", "centre")

    if "center" in options.tab_name_align:
        options.tab_name_align = str(options.tab_name_align).replace("center", "centre")

    if options.tab_side == "full" and options.tab_name_align == "edge":
        # This case does not make sense since there are two tab edges in this case.  So picking left edge.
        logger.warning(
            "Tab side 'full' and tab name align 'edge' are incompatible. Aligning card name as 'left' for 'full' tabs"
        )
        options.tab_name_align = "left"

    if options.tab_number < 1:
        logger.warning("--tab-number must be 1 or greater.  Setting to 1.")
        options.tab_number = 1

    if options.tab_side == "full" and options.tab_number != 1:
        options.tab_number = 1  # Full is 1 big tab

    if "-alternate" in options.tab_side:
        if options.tab_number != 2:
            logger.warning(
                "--tab-side with 'alternate' implies 2 tabs. Setting --tab-number to 2."
            )
        options.tab_number = 2  # alternating left and right, so override tab_number

    if "-flip" in options.tab_side:
        # for left and right tabs
        if options.tab_number != 2:
            logger.warning(
                "--tab-side with 'flip' implies 2 tabs. Setting --tab-number to 2."
            )
        options.tab_number = (
            2  # alternating left and right with a flip, so override tab_number
        )
        options.flip = True
    else:
        options.flip = False

    if options.tab_number < 3 and options.tab_serpentine:
        logger.warning("--tab-serpentine only valid if --tab-number > 2.")
        options.tab_serpentine = False

    if options.cost is None:
        options.cost = ["tab"]

    if options.set_icon is None:
        options.set_icon = ["tab"]

    if options.sleeved_thick:
        options.thickness = 3.2
        options.sleeved = True

    if options.sleeved_thin:
        options.thickness = 2.4
        options.sleeved = True

    # if notch is enabled with missing dimensions, provide defaults
    notch = options.notch or options.notch_length or options.notch_height
    if notch and not options.notch_length:
        options.notch_length = 1.5
    if notch and not options.notch_height:
        options.notch_height = 0.25

    if options.cropmarks and options.linetype == "line":
        options.linetype = "cropmarks"

    if "cropmarks" in options.linetype:
        options.cropmarks = True

    if options.expansions is None:
        # No instance given, so default to the latest Official expansions
        options.expansions = ["*"]
        if options.edition is None:
            options.edition = "latest"
    else:
        # options.expansions is a list of lists.  Reduce to single lowercase list
        options.expansions = [
            item.lower() for sublist in options.expansions for item in sublist
        ]
    if "none" in options.expansions:
        # keyword to indicate no options.  Same as --expansions without any expansions given.
        options.expansions = []

    if options.exclude_expansions:
        # options.exclude_expansions is a list of lists.  Reduce to single lowercase list
        options.exclude_expansions = [
            item.lower() for sublist in options.exclude_expansions for item in sublist
        ]

    if options.edition is None:
        # set the default
        options.edition = "all"

    if options.fan is None:
        # No instance given, so default to no Fan expansions
        options.fan = []
    else:
        # options.fan is a list of lists.  Reduce to single lowercase list
        options.fan = [item.lower() for sublist in options.fan for item in sublist]
    if "none" in options.fan:
        # keyword to indicate no options.  Same as --fan without any expansions given
        options.fan = []

    if options.only_type_any is None:
        # No instance given, so default to empty list
        options.only_type_any = []
    else:
        # options.only_type_any is a list of lists.  Reduce to single lowercase list
        options.only_type_any = list(
            set([item.lower() for sublist in options.only_type_any for item in sublist])
        )

    if options.only_type_all is None:
        # No instance given, so default to empty list
        options.only_type_all = []
    else:
        # options.only_type_any is a list of lists.  Reduce to single lowercase list
        options.only_type_all = list(
            set([item.lower() for sublist in options.only_type_all for item in sublist])
        )

    if options.group_global is None:
        options.group_global = []
    elif not any(options.group_global):
        # option given with nothing indicates all possible global groupings
        options.group_global = db.get_global_groups()[1]
    else:
        # options.group_global is a list of lists.  Reduce to single lowercase list
        options.group_global = [
            item.lower() for sublist in options.group_global for item in sublist
        ]
    # For backwards compatibility
    if options.exclude_events:
        options.group_global.append("events")
    if options.exclude_landmarks:
        options.group_global.append("landmarks")
    if options.exclude_projects:
        options.group_global.append("projects")
    if options.exclude_ways:
        options.group_global.append("ways")
    if options.exclude_traits:
        options.group_global.append("traits")
    # Remove duplicates from the list
    options.group_global = list(set(options.group_global))

    if options.tabs_only and options.label_name is None:
        # default is Avery 8867
        options.label_name = "8867"

    options.label = None
    if options.label_name is not None:
        for label in db.get_label_data()[0]:
            if options.label_name.upper() in [n.upper() for n in label["names"]]:
                options.label = label
                break

        assert options.label is not None, "Label '{}' not defined".format(
            options.label_name
        )

        # Defaults for missing values
        label = options.label
        label["paper"] = label["paper"] if "paper" in label else "LETTER"
        label["tab-only"] = label["tab-only"] if "tab-only" in label else True
        label["tab-height"] = (
            label["tab-height"] if "tab-height" in label else label["height"]
        )
        label["body-height"] = (
            label["body-height"]
            if "body-height" in label
            else label["height"] - label["tab-height"]
        )
        label["gap-vertical"] = (
            label["gap-vertical"] if "gap-vertical" in label else 0.0
        )
        label["gap-horizontal"] = (
            label["gap-horizontal"] if "gap-horizontal" in label else 0.0
        )
        label["pad-vertical"] = (
            label["pad-vertical"] if "pad-vertical" in label else 0.1
        )
        label["pad-horizontal"] = (
            label["pad-horizontal"] if "pad-horizontal" in label else 0.1
        )

        # Option Overrides when using labels
        MIN_BODY_CM_FOR_COUNT = 0.6
        MIN_BODY_CM_FOR_TEXT = 4.0
        MIN_HEIGHT_CM_FOR_VERTICAL = 5.0
        MIN_WIDTH_CM_FOR_FULL = 5.0

        options.linewidth = 0.0
        options.cropmarks = False
        options.head = "tab"
        options.tail = "none"
        options.wrapper_meta = options.pull_tab_meta = options.tent_meta = False
        options.papersize = label["paper"]
        if label["tab-only"]:
            options.tabs_only = True
        if label["body-height"] < MIN_BODY_CM_FOR_TEXT:
            # Not enough room for any text
            options.text_front = "blank"
            options.text_back = "blank"
        if label["body-height"] < MIN_BODY_CM_FOR_COUNT:
            # Not enough room for count and type
            options.count = False
            options.types = False
        if label["height"] < MIN_HEIGHT_CM_FOR_VERTICAL:
            # Not enough room to make vertical
            options.orientation = "horizontal"
        if (
            options.label["width"] - 2 * options.label["pad-horizontal"]
        ) < MIN_WIDTH_CM_FOR_FULL:
            options.tab_side = "full"
        options.label = label

    if options.wrapper_meta:
        # Same as --head=strap --tail=folder
        options.head = "strap"
        options.tail = "folder"
    if options.pull_tab_meta:
        # Same as --head=tab --tail=cover
        options.head = "tab"
        options.tail = "cover"
    if options.tent_meta:
        # Same as --head=cover --head-facing=back --head-text=back
        #         --tail=tab --tail-facing=front
        options.head = "cover"
        options.head_facing = "back"
        options.head_text = "back"
        options.tail = "tab"
        options.tail_facing = "front"
    # Flags set if there's a head wrapper, a tail wrapper, or either
    options.headWrapper = options.head in ["strap", "cover", "folder"]
    options.tailWrapper = options.tail in ["strap", "cover", "folder"]
    options.wrapper = options.headWrapper or options.tailWrapper

    # Expand --head-text and --tail-text if they refer to --front or --back
    if options.head_text == "front":
        options.head_text = options.text_front
    elif options.head_text == "back":
        options.head_text = options.text_back
    if options.tail_text == "front":
        options.tail_text = options.text_front
    elif options.tail_text == "back":
        options.tail_text = options.text_back

    return options


def parse_dimensions(dimensionsStr):
    x, y = dimensionsStr.upper().split("X", 1)
    return (float(x) * cm, float(y) * cm)


def parse_papersize(spec):
    papersize = None
    if not spec:
        if os.path.exists("/etc/papersize"):
            papersize = open("/etc/papersize").readline().upper()
        else:
            papersize = "LETTER"
    else:
        papersize = spec.upper()

    try:
        paperwidth, paperheight = getattr(pagesizes, papersize)
    except AttributeError:
        try:
            paperwidth, paperheight = parse_dimensions(papersize)
            logger.info(
                (
                    f"Using custom paper size, {paperwidth / cm:.2f}cm x {paperheight / cm:.2f}cm"
                )
            )
        except ValueError:
            paperwidth, paperheight = pagesizes.LETTER
    return paperwidth, paperheight


def parse_cardsize(spec, sleeved):
    spec = spec.upper()
    if spec == "SLEEVED" or sleeved:
        dominionCardWidth, dominionCardHeight = (9.4 * cm, 6.15 * cm)
        logger.info(
            (
                f"Using sleeved card size, {dominionCardWidth / cm:.2f}cm x {dominionCardHeight / cm:.2f}cm"
            )
        )
    elif spec in ["NORMAL", "UNSLEEVED"]:
        dominionCardWidth, dominionCardHeight = (9.1 * cm, 5.9 * cm)
        logger.info(
            (
                f"Using normal card size, {dominionCardWidth / cm:.2f}cm x{dominionCardHeight / cm:.2f}cm"
            )
        )
    else:
        dominionCardWidth, dominionCardHeight = parse_dimensions(spec)
        logger.info(
            (
                f"Using custom card size, {dominionCardWidth / cm:.2f}cm x {dominionCardHeight / cm:.2f}cm"
            )
        )
    return dominionCardWidth, dominionCardHeight



# coding=utf-8

import csv
import json
import sys

converted = []

typemap = {
    "Aktion": "Action",
    "Geld": "Treasure",
    "Fluch": "Curse",
    "Punkte": "Victory",
    "Reaktion": "Reaction",
    "Angriff": "Attack",
    "Dauer": "Duration",
    "Plündern": "Looter",
    "Ritter": "Knight",
    "Ruine": "Ruins",
    "Unterschlupf": "Shelter",
    "Reisender": "Traveller",
    "Reserve": "Reserve",
}

with open(sys.argv[1], "U") as csvfile:
    reader = csv.DictReader(csvfile)
    for row in reader:
        if not row["Typ"]:
            continue
        # row = {k: v.decode('ISO-8859-2').encode('utf-8') if type(v) == str else v for k, v in row.iteritems()}
        # print chardet.detect(row['Kartentext'])
        if "Ritter" in row["Typ"] and row["Kartenname"] != "Ritter":
            continue
        converted_row = {
            "name": row["Kartenname"],
            "cost": row["Kosten"],
            "cardset": row["Edition"],
            "description": row["Kartentext"],
            "extra": row["Lange Erklärung"],
            "types": [
                typemap[t.strip()]
                for t in row["Typ"].split("/")
                if not t.strip() == "Ritter"
            ],
            "potcost": row.get("potcost", 0),
        }
        converted.append(converted_row)
json.dump(converted, open(sys.argv[2], "wb"), indent=True)



from reportlab.lib.units import cm

from domdiv import config_options, main


def test_horizontal():
    # should be the default
    options = config_options.parse_opts([])
    assert options.orientation == "horizontal"
    main.calculate_layout(options)
    assert options.numDividersHorizontal == 2
    assert options.numDividersVertical == 3
    assert options.dividerWidth == 9.1 * cm
    assert options.labelHeight == 0.9 * cm
    assert options.dividerHeight == 5.9 * cm + options.labelHeight


def test_vertical():
    options = config_options.parse_opts(["--orientation", "vertical"])
    assert options.orientation == "vertical"
    main.calculate_layout(options)
    assert options.numDividersHorizontal == 3
    assert options.numDividersVertical == 2
    assert options.dividerWidth == 5.9 * cm
    assert options.labelHeight == 0.9 * cm
    assert options.dividerHeight == 9.1 * cm + options.labelHeight


def test_sleeved():
    options = config_options.parse_opts(["--size", "sleeved"])
    main.calculate_layout(options)
    assert options.dividerWidth == 9.4 * cm
    assert options.labelHeight == 0.9 * cm
    assert options.dividerHeight == 6.15 * cm + options.labelHeight


def test_cost():
    options = config_options.parse_opts([])
    options = config_options.clean_opts(options)
    assert options.cost == ["tab"]

    options = config_options.parse_opts(["--cost=tab"])
    options = config_options.clean_opts(options)
    assert options.cost == ["tab"]

    options = config_options.parse_opts(["--cost=body-top"])
    options = config_options.clean_opts(options)
    assert options.cost == ["body-top"]

    options = config_options.parse_opts(["--cost=hide"])
    options = config_options.clean_opts(options)
    assert options.cost == ["hide"]

    options = config_options.parse_opts(["--cost=tab", "--cost=body-top"])
    options = config_options.clean_opts(options)
    assert set(options.cost) == {"tab", "body-top"}


def test_set_icon():
    options = config_options.parse_opts([])
    options = config_options.clean_opts(options)
    assert options.set_icon == ["tab"]

    options = config_options.parse_opts(["--set-icon=tab"])
    options = config_options.clean_opts(options)
    assert options.set_icon == ["tab"]

    options = config_options.parse_opts(["--set-icon=body-top"])
    options = config_options.clean_opts(options)
    assert options.set_icon == ["body-top"]

    options = config_options.parse_opts(["--set-icon=hide"])
    options = config_options.clean_opts(options)
    assert options.set_icon == ["hide"]

    options = config_options.parse_opts(["--set-icon=tab", "--set-icon=body-top"])
    options = config_options.clean_opts(options)
    assert set(options.set_icon) == {"tab", "body-top"}



from __future__ import print_function

import contextlib
import os
import shutil
import unicodedata

import pytest

from domdiv import cards as domdiv_cards
from domdiv import config_options, db, main


@pytest.fixture
def rmtestcardb(request):
    def rmd():
        testcardb_dir = os.path.join(str(request.config.rootdir), "tools/card_db")
        if os.path.exists(testcardb_dir):
            print("removing {}".format(testcardb_dir))
            shutil.rmtree(testcardb_dir)

    request.addfinalizer(rmd)


def test_cardread():
    # we test the number of cards only to make sure it doesn't get changed
    # inadvertently by unrelated changes
    num_cards_expected = 1018

    options = config_options.parse_opts([])
    cards = db.read_card_data(options)
    assert len(cards) == num_cards_expected
    valid_cardsets = {
        "base",
        "dominion1stEdition",
        "dominion1stEditionRemoved",
        "dominion2ndEdition",
        "dominion2ndEditionUpgrade",
        "intrigue1stEdition",
        "intrigue1stEditionRemoved",
        "intrigue2ndEdition",
        "intrigue2ndEditionUpgrade",
        "seaside1stEdition",
        "seaside1stEditionRemoved",
        "seaside2ndEdition",
        "seaside2ndEditionUpgrade",
        "alchemy",
        "prosperity1stEdition",
        "prosperity1stEditionRemoved",
        "prosperity2ndEdition",
        "prosperity2ndEditionUpgrade",
        "cornucopia extras",
        "cornucopia1stEdition",
        "cornucopia1stEditionRemoved",
        "cornucopia2ndEditionUpgrade",
        "cornucopiaAndGuilds2ndEdition",
        "hinterlands1stEdition",
        "hinterlands1stEditionRemoved",
        "hinterlands2ndEdition",
        "hinterlands2ndEditionUpgrade",
        "dark ages",
        "dark ages extras",
        "guilds1stEdition",
        "guilds1stEditionRemoved",
        "guilds2ndEditionUpgrade",
        "guilds-bigbox2-de",
        "adventures",
        "adventures extras",
        "empires",
        "empires extras",
        "nocturne",
        "nocturne extras",
        "plunder",
        "promo",
        "promo-bigbox2-de",
        "renaissance",
        "menagerie",
        "extras",
        "animals",
        "allies",
    }
    for c in cards:
        assert isinstance(c, domdiv_cards.Card)
        assert c.cardset_tag in valid_cardsets

    # Option modified card count
    options = config_options.parse_opts(
        ["--no-trash", "--curse10", "--start-decks", "--include-blanks", "7"]
    )
    options = config_options.clean_opts(options)
    options.data_path = "."
    cards = db.read_card_data(options)
    # Total delta cards is +28 from
    #      Trash:       -1 * 3 sets = -3
    #      Curse:       +2 * 4 sets = +8
    #      Start Decks: +4 * 4 sets = +16
    #      Blanks:      +7          = +7
    assert len(cards) == num_cards_expected + 28


@pytest.mark.parametrize("lang", db.get_languages("card_db"))
def test_languages_db(lang):
    print("checking " + lang)
    # for now, just test that they load
    options = config_options.parse_opts(["--language", lang])
    options.data_path = "."
    cards = db.read_card_data(options)
    assert cards, '"{}" cards did not read properly'.format(lang)
    cards = main.add_card_text(cards, "en_us")
    cards = main.add_card_text(cards, lang)
    if lang == "it":
        assert "Maledizione" in [card.name for card in cards]
    elif lang == "de":
        assert "Fluch" in [card.name for card in cards]


@contextlib.contextmanager
def change_cwd(d):
    curdir = os.getcwd()
    try:
        os.chdir(d)
        yield
    finally:
        os.chdir(curdir)


def test_only_type():
    options = config_options.parse_opts(
        [
            "--expansions",
            "base",
            "alchemy",
            "--include-blanks",
            "5",
            "--only-type-any",
            "blank",
            "curse",
            "--only-type-all",
            "attack",
            "action",
        ]
    )
    options = config_options.clean_opts(options)
    options.data_path = "."
    cards = db.read_card_data(options)
    cards = main.filter_sort_cards(cards, options)
    # Total 8 from
    #      Blank:         +5 added in options
    #      Curse:         +1 from base
    #      Action Attack: +2 from Alchemy
    print(cards)
    assert len(cards) == 8


def test_expansion():
    # test that we can use --expansion or
    # --expansions, that we can have multiple
    # items with a single flag, that * syntax
    # works, that we can use either the
    # cardset tag or name, and that capitalization
    # doesn't matter
    options = config_options.parse_opts(
        [
            "--expansion",
            "advEntUres",
            "dominion 2nd*",
            "--expansions=intrigue1stEdition",
        ]
    )
    options = config_options.clean_opts(options)
    options.data_path = "."
    cards = db.read_card_data(options)
    cards = main.filter_sort_cards(cards, options)
    card_sets = set(x.cardset.lower() for x in cards)
    assert card_sets == {
        "adventures",
        "dominion 2nd edition",
        "dominion 2nd edition upgrade",
        "intrigue 1st edition",
    }


def test_exclude_expansion():
    # test that we can use --exclude-expansion or
    # --exclude-expansions, that we can have multiple
    # items with a single flag, that * syntax
    # works, that we can use either the
    # cardset tag or name, and that capitalization
    # doesn't matter
    options = config_options.parse_opts(
        [
            "--expansions",
            "adventures",
            "dominion*",
            "intrigue*",
            "--exclude-expansions",
            "dominiOn1stEditIon",
            "intrigue 2nd*",
            "--exclude-expansion",
            "dominion 2nd edition",
        ]
    )
    options = config_options.clean_opts(options)
    options.data_path = "."
    cards = db.read_card_data(options)
    cards = main.filter_sort_cards(cards, options)
    card_sets = set(x.cardset.lower() for x in cards)
    assert card_sets == {
        "adventures",
        "dominion 1st edition removed",
        "dominion 2nd edition upgrade",
        "intrigue 1st edition",
        "intrigue 1st edition removed",
    }


def test_expansion_description_card_order():
    # test that the expansions cards lists cards
    # in alphabetical order, like they are printed,
    # and that accents don't matter
    options = config_options.parse_opts(
        [
            "--expansions",
            "hinterlands1stEdition",
            "--expansion-dividers",
            "--language",
            "fr",
            "--only-type-any",
            "Expansion",
        ]
    )
    options = config_options.clean_opts(options)
    options.data_path = "."
    cards = db.read_card_data(options)
    cards = main.filter_sort_cards(cards, options)
    card_names = [c.strip() for c in cards[0].description.split("|")]
    # The 26 french card names of the Hinterlands expansion should be sorted as if no accent
    assert len(card_names) == 26
    assert card_names == sorted(
        card_names,
        key=lambda s: "".join(
            c
            for c in unicodedata.normalize("NFD", s)
            if unicodedata.category(c) != "Mn"
        ),
    )



import glob
import os

from domdiv.tools import bgg_release, update_language

DOIT_CONFIG = {"default_tasks": ["build"]}


def glob_no_dirs(spec):
    return [fname for fname in glob.glob(spec) if os.path.isfile(fname)]


def task_compile_requirements():
    return {
        "file_dep": ["pyproject.toml"],
        "actions": [
            "pip-compile -U --no-emit-index-url --resolver=backtracking pyproject.toml",
        ],
        "targets": ["requirements.txt"],
    }


def task_update_languages():
    files = glob.glob("card_db_src/**/*.json") + glob.glob("card_db_src/*.json")
    return {
        "file_dep": files
        + ["src/domdiv/tools/update_language.py", "src/domdiv/tools/common.py"],
        "actions": [lambda: update_language.main("card_db_src", "src/domdiv/card_db")],
        "targets": [
            os.path.join(
                "src",
                "domdiv",
                "card_db",
                os.path.sep.join(fname.split(os.path.sep)[1:]),
            )
            for fname in files
        ],
        "clean": True,
    }


def task_build():
    files = [
        fname
        for fname in glob_no_dirs("src/domdiv/**/*")
        + glob.glob("card_db_src/**/*.json" + "pyproject.toml")
        if os.path.isfile(fname)
    ]
    return {
        "file_dep": files,
        "task_dep": ["update_languages"],
        "actions": ["pip install -e .[dev]", "python -m build"],
    }


def task_make_bgg_release():
    return {"actions": [lambda: bgg_release.make_bgg_release()]}


def task_test():
    files = glob_no_dirs("src/domdiv/**")
    return {"file_dep": files, "actions": ["pip install -e .[dev]", "pytest"]}



###########################################################################
# This file provides maintenance on the various language files
# 1. Create new "xx/cards_xx.json" files that have entries ordered as:
#     a. the card_tag entries in "cards_db.json"
#     b. the group_tag entries as found in "cards_db.json"
#     c. the super group entries (grouping across all expansions"
#     d. any unused entries existing in the file (assumed to be work in progress)
#
# 2. Create new "sets_db.json" and "xx/cards_xx.json" with entries sorted alphabetically
#
# All output is in the designated output directory.  Original files are not overwritten.
###########################################################################

import argparse
import collections
import gzip
import os
from shutil import copyfile

from domdiv.tools.common import (
    LANGUAGE_DEFAULT,
    LANGUAGE_XX,
    check_compressed_json_change,
    get_json_data,
    get_languages,
    load_card_data,
    load_language_cards,
    multikeysort,
    write_data,
    write_language_cards,
)

VALID_CARD_FIELD_NAMES = {"description", "extra", "name"}


def main(card_db_dir, output_dir):
    ###########################################################################
    # Get all the languages, and place the default language first in the list
    ###########################################################################
    languages = get_languages(card_db_dir)
    print("Languages:")
    print(languages)
    print()

    ###########################################################################
    #  Make sure the directories exist to hold the output
    ###########################################################################

    #  main output directory
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)

    #  each language directory
    for lang in languages:
        #  Make sure the directory is there to hold the file
        lang_dir = os.path.join(output_dir, lang)
        if not os.path.exists(lang_dir):
            os.makedirs(lang_dir)

    ###########################################################################
    #  Get the types_db information
    #  Store in a list in the order found in types[]. Ordered by card_type
    #  1. card_tags, 2. group_tags, 3. super groups
    ###########################################################################
    type_parts = set()

    # Get the card data
    type_data = get_json_data(os.path.join(card_db_dir, "types_db.json"))
    types = set(tuple(t["card_type"]) for t in type_data)
    assert len(type_data) == len(types), f"{len(type_data)} != {len(types)}"

    # Sort the cards by cardset_tags, then card_tag
    sorted_type_data = multikeysort(type_data, ["card_type"])

    write_data(sorted_type_data, os.path.join(output_dir, "types_db.json"))

    type_parts = list(set().union(*[set(t["card_type"]) for t in sorted_type_data]))
    type_parts.sort()
    print("Unique Types:")
    print(type_parts)
    print()

    ###########################################################################
    #  Get the labels_db information
    #  Store in a list in the order found.
    ###########################################################################
    all_labels = []

    # Get the card data
    label_data = get_json_data(os.path.join(card_db_dir, "labels_db.json"))

    all_labels = list(set().union(*[set(label["names"]) for label in label_data]))

    write_data(label_data, os.path.join(output_dir, "labels_db.json"))

    all_labels.sort()
    print("Labels: ")
    print(all_labels)
    print()
    ###########################################################################
    # Fix up all the xx/types_xx.json files
    # Place entries in alphabetical order
    # If entries don't exist:
    #    If the default language, set from information in the "types_db.json" file,
    #    If not the default language, set based on information from the default language.
    # Lastly, keep any extra entries that are not currently used, just in case needed
    #    in the future or is a work in progress.
    ###########################################################################
    for lang in languages:
        lang_file = "types_" + lang + ".json"
        fname = os.path.join(card_db_dir, lang, lang_file)
        if os.path.isfile(fname):
            lang_type_data = get_json_data(fname)
        else:
            lang_type_data = {}

        for t in sorted(type_parts):
            if t not in lang_type_data:
                if lang == LANGUAGE_DEFAULT:
                    lang_type_data[t] = t
                    lang_type_default = lang_type_data
                else:
                    lang_type_data[t] = lang_type_default[t]
        write_data(lang_type_data, os.path.join(output_dir, lang, lang_file))

        if lang == LANGUAGE_DEFAULT:
            lang_type_default = lang_type_data  # Keep for later languages

    sorted_card_data = load_card_data(card_db_dir)
    seen = set()
    for c in sorted_card_data:
        if c["card_tag"] in seen:
            raise RuntimeError(f"Duplicate card detected: {c['card_tag']}")
        seen.add(c["card_tag"])
    groups = set(card["group_tag"] for card in sorted_card_data if "group_tag" in card)
    super_groups = set(["events", "landmarks", "projects"])

    write_data(sorted_card_data, os.path.join(output_dir, "cards_db.json"))

    # maintain the sorted order, but expand with groups and super_groups
    cards = [c["card_tag"] for c in sorted_card_data]
    cards.extend(sorted(groups))
    cards.extend(sorted(super_groups))

    print("Cards:")
    print(cards)
    print()

    ###########################################################################
    # Fix up all the cards_xx.json files
    # Place entries in the same order as given in "cards_db.json".
    # If entries don't exist:
    #    If the default language, set base on information in the "cards_db.json" file,
    #    If not the default language, set based on information from the default language.
    # Lastly, keep any extra entries that are not currently used, just in case needed
    #    in the future or is a work in progress.
    ###########################################################################

    for lang in languages:
        #  contruct the cards json file name
        lang_data = load_language_cards(lang, card_db_dir)

        sorted_lang_data = collections.OrderedDict()
        for card_tag in cards:
            lang_card = lang_data.get(card_tag)

            # print(f'looking at {card_tag}: {lang_card}')
            if not lang_card or lang == LANGUAGE_XX:
                #  Card is missing, need to add it
                lang_card = {}
                if lang == LANGUAGE_DEFAULT:
                    lang_default = lang_data
                    extra_fields = set(lang_card) - VALID_CARD_FIELD_NAMES
                    assert (
                        len(extra_fields) == 0
                    ), f"invalid extra field names for {card_tag} ({lang}): {extra_fields}"
                else:
                    #  All other languages should get the default languages' text
                    lang_card = lang_default.get(card_tag, {}).copy()
            elif lang != LANGUAGE_DEFAULT:
                # Card exists, figure out what needs updating
                extra_fields = set(lang_card) - VALID_CARD_FIELD_NAMES
                assert (
                    len(extra_fields) == 0
                ), f"invalid extra field names for {card_tag} ({lang}): {extra_fields}"
                lang_card.update(
                    {
                        field: value
                        for field, value in lang_default.get(card_tag, {}).items()
                        if field not in lang_card
                    }
                )
            if lang_card:
                sorted_lang_data[card_tag] = lang_card
        unused = set(lang_data) - set(sorted_lang_data)
        print(
            f"unused in {lang}: {len(unused)}, used: {len(set(lang_data) & set(sorted_lang_data))}"
        )
        print(unused)
        # Now keep any unused values just in case needed in the future
        for card_tag in sorted(unused):
            lang_card = lang_data.get(card_tag)
            if lang_card:
                lang_card["notes"] = ["This card is currently not used."]
                sorted_lang_data[card_tag] = lang_card

        write_language_cards(sorted_lang_data, lang, output_dir)

        if lang == LANGUAGE_DEFAULT:
            lang_default = lang_data  # Keep for later languages

    ###########################################################################
    # Fix up the sets_db.json file
    # Place entries in alphabetical order
    ###########################################################################
    lang_file = "sets_db.json"
    set_data = get_json_data(os.path.join(card_db_dir, lang_file))

    write_data(set_data, os.path.join(output_dir, lang_file))

    print("Sets:")
    print(set(set_data))
    print()

    ###########################################################################
    # Fix up all the xx/sets_xx.json files
    # Place entries in alphabetical order
    # If entries don't exist:
    #    If the default language, set from information in the "sets_db.json" file,
    #    If not the default language, set based on information from the default language.
    ###########################################################################
    for lang in languages:
        lang_file = "sets_" + lang + ".json"
        fname = os.path.join(card_db_dir, lang, lang_file)
        if os.path.isfile(fname):
            lang_set_data = get_json_data(fname)
        else:
            lang_set_data = {}

        for s in sorted(set_data):
            if s not in lang_set_data:
                lang_set_data[s] = {}
                if lang == LANGUAGE_DEFAULT:
                    lang_set_data[s]["set_name"] = s.title()
                    lang_set_data[s]["text_icon"] = set_data[s]["text_icon"]
                    if "short_name" in set_data[s]:
                        lang_set_data[s]["short_name"] = set_data[s]["short_name"]
                    if "set_text" in set_data[s]:
                        lang_set_data[s]["set_text"] = set_data[s]["set_text"]
                else:
                    lang_set_data[s]["set_name"] = lang_default[s]["set_name"]
                    lang_set_data[s]["text_icon"] = lang_default[s]["text_icon"]
                    if "short_name" in lang_default[s]:
                        lang_set_data[s]["short_name"] = lang_default[s]["short_name"]
                    if "set_text" in lang_default[s]:
                        lang_set_data[s]["set_text"] = lang_default[s]["set_text"]
            else:
                if lang != LANGUAGE_DEFAULT:
                    for x in lang_default[s]:
                        if x not in lang_set_data[s] and x != "used":
                            lang_set_data[s][x] = lang_default[s][x]

        if lang == LANGUAGE_DEFAULT:
            lang_default = lang_set_data  # Keep for later languages

        write_data(lang_set_data, os.path.join(output_dir, lang, lang_file))

    ###########################################################################
    # bonuses_xx files
    ###########################################################################
    for lang in languages:
        # Special case for xx.  Reseed from default language
        fromLanguage = lang
        if lang == LANGUAGE_XX:
            fromLanguage = LANGUAGE_DEFAULT

        with open(
            os.path.join(card_db_dir, fromLanguage, "bonuses_" + fromLanguage + ".json")
        ) as f:
            data = f.read()
            output_fname = os.path.join(output_dir, lang, f"bonuses_{lang}.json.gz")
            if check_compressed_json_change(output_fname, data):
                with gzip.open(
                    output_fname,
                    "wt",
                    encoding="utf-8",
                ) as fout:
                    fout.write(data)

    ###########################################################################
    # translation.txt
    ###########################################################################
    copyfile(
        os.path.join(card_db_dir, "translation.md"),
        os.path.join(output_dir, "translation.md"),
    )

    # Since xx is the starting point for new translations,
    # make sure xx has the latest copy of translation.txt
    copyfile(
        os.path.join(card_db_dir, LANGUAGE_XX, "translation.txt"),
        os.path.join(output_dir, LANGUAGE_XX, "translation.txt"),
    )


def run():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "card_db_dir",
        help="directory of card data (usually card_db_src at the top level of the distribution)",
    )
    parser.add_argument(
        "output_dir", help="directory for output data (usually src/domdiv/card_db)"
    )
    args = parser.parse_args()
    main(args.card_db_dir, args.output_dir)


if __name__ == "__main__":
    run()



from domdiv import config_options, main


####################
# Card Text and Tab Default Test
####################
def test_text_tabs_default():
    # should be the default
    options = config_options.parse_opts([])
    assert options.text_front == "card"
    assert options.text_back == "rules"
    assert options.tab_name_align == "left"
    assert options.tab_side == "right-alternate"
    main.calculate_layout(options)
    assert options.tab_name_align == "left"


####################
# Card Text Tests
####################


def test_text_card_rules():
    options = config_options.parse_opts(["--front", "card", "--back", "rules"])
    assert options.text_front == "card"
    assert options.text_back == "rules"


def test_text_card_blank():
    options = config_options.parse_opts(["--front", "card", "--back", "blank"])
    assert options.text_front == "card"
    assert options.text_back == "blank"


def test_text_card_card():
    options = config_options.parse_opts(["--front", "card", "--back", "card"])
    assert options.text_front == "card"
    assert options.text_back == "card"


def test_text_card_none():
    options = config_options.parse_opts(["--front", "card", "--back", "none"])
    assert options.text_front == "card"
    assert options.text_back == "none"


def test_text_rules_rules():
    options = config_options.parse_opts(["--front", "rules", "--back", "rules"])
    assert options.text_front == "rules"
    assert options.text_back == "rules"


def test_text_rules_blank():
    options = config_options.parse_opts(["--front", "rules", "--back", "blank"])
    assert options.text_front == "rules"
    assert options.text_back == "blank"


def test_text_rules_card():
    options = config_options.parse_opts(["--front", "rules", "--back", "card"])
    assert options.text_front == "rules"
    assert options.text_back == "card"


def test_text_rules_none():
    options = config_options.parse_opts(["--front", "rules", "--back", "none"])
    assert options.text_front == "rules"
    assert options.text_back == "none"


def test_text_blank_rules():
    options = config_options.parse_opts(["--front", "blank", "--back", "rules"])
    assert options.text_front == "blank"
    assert options.text_back == "rules"


def test_text_blank_blank():
    options = config_options.parse_opts(["--front", "blank", "--back", "blank"])
    assert options.text_front == "blank"
    assert options.text_back == "blank"


def test_text_blank_card():
    options = config_options.parse_opts(["--front", "blank", "--back", "card"])
    assert options.text_front == "blank"
    assert options.text_back == "card"


def test_text_blank_none():
    options = config_options.parse_opts(["--front", "blank", "--back", "none"])
    assert options.text_front == "blank"
    assert options.text_back == "none"


####################
# Card Tab Tests
####################
# --tab_name_align left


def test_tab_left_left():
    options = config_options.parse_opts(
        ["--tab-name-align", "left", "--tab-side", "left"]
    )
    assert options.tab_name_align == "left"
    assert options.tab_side == "left"
    main.calculate_layout(options)
    assert options.tab_name_align == "left"
    assert options.tab_side == "left"


def test_tab_left_right():
    options = config_options.parse_opts(
        ["--tab-name-align", "left", "--tab-side", "right"]
    )
    assert options.tab_name_align == "left"
    assert options.tab_side == "right"
    main.calculate_layout(options)
    assert options.tab_name_align == "left"
    assert options.tab_side == "right"


def test_tab_left_leftalt():
    options = config_options.parse_opts(
        ["--tab-name-align", "left", "--tab-side", "left-alternate"]
    )
    assert options.tab_name_align == "left"
    assert options.tab_side == "left-alternate"
    main.calculate_layout(options)
    assert options.tab_name_align == "left"
    assert options.tab_side == "left-alternate"


def test_tab_left_rightalt():
    options = config_options.parse_opts(
        ["--tab-name-align", "left", "--tab-side", "right-alternate"]
    )
    assert options.tab_name_align == "left"
    assert options.tab_side == "right-alternate"
    main.calculate_layout(options)
    assert options.tab_name_align == "left"
    assert options.tab_side == "right-alternate"


def test_tab_left_full():
    options = config_options.parse_opts(
        ["--tab-name-align", "left", "--tab-side", "full"]
    )
    assert options.tab_name_align == "left"
    assert options.tab_side == "full"
    main.calculate_layout(options)
    assert options.tab_name_align == "left"
    assert options.tab_side == "full"

    # --tab_name_align right


def test_tab_right_left():
    options = config_options.parse_opts(
        ["--tab-name-align", "right", "--tab-side", "left"]
    )
    assert options.tab_name_align == "right"
    assert options.tab_side == "left"
    main.calculate_layout(options)
    assert options.tab_name_align == "right"
    assert options.tab_side == "left"


def test_tab_right_right():
    options = config_options.parse_opts(
        ["--tab-name-align", "right", "--tab-side", "right"]
    )
    assert options.tab_name_align == "right"
    assert options.tab_side == "right"
    main.calculate_layout(options)
    assert options.tab_name_align == "right"
    assert options.tab_side == "right"


def test_tab_right_leftalt():
    options = config_options.parse_opts(
        ["--tab-name-align", "right", "--tab-side", "left-alternate"]
    )
    assert options.tab_name_align == "right"
    assert options.tab_side == "left-alternate"
    main.calculate_layout(options)
    assert options.tab_name_align == "right"
    assert options.tab_side == "left-alternate"


def test_tab_right_rightalt():
    options = config_options.parse_opts(
        ["--tab-name-align", "right", "--tab-side", "right-alternate"]
    )
    assert options.tab_name_align == "right"
    assert options.tab_side == "right-alternate"
    main.calculate_layout(options)
    assert options.tab_name_align == "right"
    assert options.tab_side == "right-alternate"


def test_tab_right_full():
    options = config_options.parse_opts(
        ["--tab-name-align", "right", "--tab-side", "full"]
    )
    assert options.tab_name_align == "right"
    assert options.tab_side == "full"
    main.calculate_layout(options)
    assert options.tab_name_align == "right"
    assert options.tab_side == "full"


# --tab_name_align edge


def test_tab_edge_left():
    options = config_options.parse_opts(
        ["--tab-name-align", "edge", "--tab-side", "left"]
    )
    assert options.tab_name_align == "edge"
    assert options.tab_side == "left"
    main.calculate_layout(options)
    assert options.tab_name_align == "edge"
    assert options.tab_side == "left"


def test_tab_edge_right():
    options = config_options.parse_opts(
        ["--tab-name-align", "edge", "--tab-side", "right"]
    )
    assert options.tab_name_align == "edge"
    assert options.tab_side == "right"
    main.calculate_layout(options)
    assert options.tab_name_align == "edge"
    assert options.tab_side == "right"


def test_tab_edge_leftalt():
    options = config_options.parse_opts(
        ["--tab-name-align", "edge", "--tab-side", "left-alternate"]
    )
    assert options.tab_name_align == "edge"
    assert options.tab_side == "left-alternate"
    main.calculate_layout(options)
    assert options.tab_name_align == "edge"
    assert options.tab_side == "left-alternate"


def test_tab_edge_rightalt():
    options = config_options.parse_opts(
        ["--tab-name-align", "edge", "--tab-side", "right-alternate"]
    )
    assert options.tab_name_align == "edge"
    assert options.tab_side == "right-alternate"
    main.calculate_layout(options)
    assert options.tab_name_align == "edge"
    assert options.tab_side == "right-alternate"


def test_tab_edge_full():
    options = config_options.parse_opts(
        ["--tab-name-align", "edge", "--tab-side", "full"]
    )
    assert options.tab_name_align == "edge"
    assert options.tab_side == "full"
    options = config_options.clean_opts(options)
    main.calculate_layout(options)
    assert options.tab_name_align == "left"  # special check for odd condition
    assert options.tab_side == "full"

    # --tab_name_align centre


def test_tab_centre_left():
    options = config_options.parse_opts(
        ["--tab-name-align", "centre", "--tab-side", "left"]
    )
    assert options.tab_name_align == "centre"
    assert options.tab_side == "left"
    main.calculate_layout(options)
    assert options.tab_name_align == "centre"
    assert options.tab_side == "left"


def test_tab_centre_right():
    options = config_options.parse_opts(
        ["--tab-name-align", "centre", "--tab-side", "right"]
    )
    assert options.tab_name_align == "centre"
    assert options.tab_side == "right"
    main.calculate_layout(options)
    assert options.tab_name_align == "centre"
    assert options.tab_side == "right"


def test_tab_centre_leftalt():
    options = config_options.parse_opts(
        ["--tab-name-align", "centre", "--tab-side", "left-alternate"]
    )
    assert options.tab_name_align == "centre"
    assert options.tab_side == "left-alternate"
    main.calculate_layout(options)
    assert options.tab_name_align == "centre"
    assert options.tab_side == "left-alternate"


def test_tab_centre_rightalt():
    options = config_options.parse_opts(
        ["--tab-name-align", "centre", "--tab-side", "right-alternate"]
    )
    assert options.tab_name_align == "centre"
    assert options.tab_side == "right-alternate"
    main.calculate_layout(options)
    assert options.tab_name_align == "centre"
    assert options.tab_side == "right-alternate"


def test_tab_centre_full():
    options = config_options.parse_opts(
        ["--tab-name-align", "centre", "--tab-side", "full"]
    )
    assert options.tab_name_align == "centre"
    assert options.tab_side == "full"
    main.calculate_layout(options)
    assert options.tab_name_align == "centre"
    assert options.tab_side == "full"


# --tab_name_align center.  Just do one since this is an alias to centre


def test_tab_center_left():
    options = config_options.parse_opts(
        ["--tab-name-align", "center", "--tab-side", "left"]
    )
    assert options.tab_name_align == "center"
    assert options.tab_side == "left"
    options = config_options.clean_opts(options)
    main.calculate_layout(options)
    assert options.tab_name_align == "centre"  # check for change in value
    assert options.tab_side == "left"
